#!/usr/bin/env python
# coding=utf-8
"""
Rucio Consistency Monitor Service class developed on top of the native WMCore
Service APIs providing custom output and handling as necessary for the
CMS Workload Management system
"""

from __future__ import division, print_function, absolute_import
from future import standard_library

from urllib.parse import urlencode

import json
import logging

from WMCore.Services.Service import Service
from Utils.Utilities import decodeBytesToUnicode


standard_library.install_aliases()


class RucioConMon(Service):
    """
    API for dealing with retrieving information from Rucio Consistency Monitor
    """

    def __init__(self, url, logger=None, configDict=None):
        """
        Init method for the RucioConMon Class
        """
        configDict = configDict or {}
        configDict.setdefault('endpoint', url)
        configDict.setdefault('cacheduration', 1)  # in hours
        configDict.setdefault('accept_type', 'application/json')
        configDict.setdefault('content_type', 'application/json')
        configDict['logger'] = logger if logger else logging.getLogger()
        super(RucioConMon, self).__init__(configDict)
        self['logger'].debug("Initializing RucioConMon with url: %s", self['endpoint'])

    def _getResult(self, uri, callname="", clearCache=False, args=None):
        """
        Either fetch data from the cache file or query the data-service
        :param uri: The endpoint uri
        :param callname: alias for caller function
        :param clearCache: parameter to control the cache behavior
        :param args: additional parameters to HTTP request call
        :return:    A dictionary
        """

        # NOTE: Unlike the common case we are not using the callname for building
        #       apiUrl. we  are using it only for giving proper name of the cache
        #       file.
        # NOTE: The 'callname' should not contain '/', otherwise the cache is
        #       tried to be created with a a non existing subdirectory structure

        cachedApi = "%s.json" % callname
        # apiUrl = '%s?json&preset=%s' % (uri, callname)
        apiUrl = uri

        self['logger'].debug('Fetching data from %s, with args %s', apiUrl, args)
        if args:
            apiUrl = "%s&%s" % (apiUrl, urlencode(args, doseq=True))

        if clearCache:
            self.clearCache(cachedApi, args)
        with self.refreshCache(cachedApi, apiUrl, decoder=True, binary=False) as istream:
            return json.load(istream)

    def _getResultZipped(self, uri, callname="", clearCache=True):
        """
        This method retrieves gzipped content, instead of the standard json format.
        :param uri: The endpoint uri
        :param callname: alias for caller function
        :param clearCache: parameter to control the cache behavior
        :return: yields a single record from the data retrieved
        """
        cachedApi = callname
        if clearCache:
            self.clearCache(cachedApi)

        with self.refreshCache(cachedApi, uri, decoder=False, binary=True) as istream:
            for line in istream:
                line = decodeBytesToUnicode(line).replace("\n", "")
                yield line

    def getRSEStats(self):
        """
        Gets the latest statistics from the RucioConMon, together with the last
        update timestamps for all RSEs known to CMS Rucio
        :return: A dictionary
        """
        uri = "stats"
        rseStats = self._getResult(uri, callname='stats')
        return rseStats

    def getRSEUnmerged(self, rseName, zipped=False):
        """
        Gets the list of all unmerged files in an RSE
        :param rseName: The RSE whose list of unmerged files to be retrieved
        :param zipped:  If True the interface providing the zipped lists will be called
        :return: a generator of unmerged files for the RSE in question
        """
        # NOTE: The default API provided by Rucio Consistency Monitor is in a form of a
        #       zipped file/stream. Currently we are using the newly provided json API
        #       But in in case we figure out the data is too big we may need to
        #       implement the method with the zipped API and use disc cache for
        #       reading/streaming from file. This will prevent any set arithmetic
        #       in the future.
        if zipped:
            uri = "files?rse=%s&format=raw" % rseName
            callname = '{}.zipped'.format(rseName)
            rseUnmerged = self._getResultZipped(uri, callname=callname, clearCache=True)
        else:
            uri = "files?rse=%s&format=json" % rseName
            callname = '{}.json'.format(rseName)
            rseUnmerged = self._getResult(uri, callname=callname)
        # now lazily return items
        for item in rseUnmerged:
            yield item



#!/usr/bin/env python

"""
_GetForJobSplittingByID_

MySQL implementation of File.GetForJobSplittingByID
"""

from WMCore.Database.DBFormatter import DBFormatter
from WMCore.WMBS.MySQL.Files.GetByID import GetByID


class GetForJobSplittingByID(GetByID):
    sql = """SELECT id, lfn, filesize, events, first_event, merged, MIN(run) AS minrun
             FROM wmbs_file_details wfd
             LEFT OUTER JOIN wmbs_file_runlumi_map wfr ON wfr.fileid = wfd.id
             WHERE id = :fileid"""

    def formatBulkDict(self, result):
        """
        _formatBulkDict_

        Formats a whole list of dictionaries
        """

        formattedResult = {}
        listOfDicts = DBFormatter.formatDict(self, result)

        for entry in listOfDicts:
            tmpDict = {}
            tmpDict["id"] = int(entry["id"])
            tmpDict["lfn"] = entry["lfn"]
            tmpDict["events"] = int(entry["events"])
            tmpDict["first_event"] = int(entry["first_event"])
            tmpDict['minrun'] = entry.get('minrun', None)
            if "size" in entry:
                tmpDict["size"] = int(entry["size"])
            else:
                tmpDict["size"] = int(entry["filesize"])
                del entry["filesize"]
            formattedResult[tmpDict['id']] = tmpDict

        return formattedResult



from __future__ import print_function, division

UNIFIED_CONFIG = {
    "site_for_overflow": {
        "value": ["nono_T0_CH_CERN",
                  "T2_CH_CERN_HLT"],
        "description": "The sites that we set to overflow and require a specific treatment"
    },
    "overflow_pressure": {
        "value": 0.5,
        "description": "The ratio pending/running over which to consider overflowing"
    },
    "DDM_buffer_level": {
        "value": 0.8,
        "description": "The fraction of the DDM quota we are allowed to use"
    },
    "sites_banned": {
        "value": ["T2_CH_CERN_AI",
                  "NONO_T0_CH_CERN",
                  "T2_TH_CUNSTDA",
                  "NONO_T2_US_Vanderbilt",
                  "T2_EE_Estonia",
                  "T2_UA_KIPT"
                  ],
        "description": "The sites that are banned from production"
    },
    "sites_auto_approve": {
        "value": ["T0_CH_CERN_MSS", "T1_FR_CCIN2P3_MSS"],
        "description": "The sites we can autoapprove tape request to"
    },
    "sites_space_override": {
        "value": {
        },
        "description": "Over-ride the available space at a phedex end point"
    },
    "sites_with_goodIO": {
        "value": ["T2_DE_DESY", "T2_DE_RWTH", "T2_ES_CIEMAT", "T2_FR_GRIF_LLR", "T2_FR_GRIF_IRFU", "T2_FR_IPHC",
                  "T2_FR_CCIN2P3", "T2_IT_Bari", "T2_IT_Legnaro", "T2_IT_Pisa", "T2_IT_Rome", "T2_UK_London_Brunel",
                  "T2_UK_London_IC", "T2_US_Caltech", "T2_US_MIT", "T2_US_Nebraska", "T2_US_Purdue", "T2_US_UCSD",
                  "T2_US_Wisconsin", "T2_US_Florida", "T2_BE_IIHE", "T2_EE_Estonia", "T2_PL_Swierk", "T2_CH_CERN",
                  "T2_CH_CERN_HLT"],
        "description": "The sites identified as having good storage badnwidth"
    },
    "max_cpuh_block": {
        "value": 40000000,
        "description": "Value of CPUh above which a wf is blocked from assigning"
    },
    "block_repositionning": {
        "value": True,
        "description": "Whether or not to retransfer block from WQE without location"
    },
    "allowed_bypass": {
        "description": "Who is allowed to bypass and force complete",
        "value": [["vlimant", "vlimant@cern.ch"],
                  ["prozober", "paola.katherine.rozo.bernal@cern.ch"],
                  ["mcremone", "matteoc@fnal.gov"]]
    },
    "max_tail_priority": {
        "value": 5,
        "description": "Number of workflow to increase the priority of at a time"
    },
    "injection_delay_threshold": {
        "value": 50,
        "description": "Number of days after wich to increase the priority of a workflow"
    },
    "delay_priority_increase": {
        "value": 10000,
        "description": "Priority from original increase per week over the delay threshold"
    },
    "injection_delay_priority": {
        "value": 75000,
        "description": "Priority above which we can increase the priority of a workflow after running too long"
    },

    "max_force_complete": {
        "value": 10,
        "description": "Number of workflow that can be forced complete at a time"
    },
    "max_per_round": {
        "description": "limitation on the number of wf to process per module",
        "value": {
            "transferor": None,
            "assignor": None,
            "closor": None,
            "checkor": None,
            "completor": None
        }
    },
    "default_fraction_pass": {
        "value": 1.0,
        "description": "completion fraction above which to announce dataset"
    },
    "pattern_fraction_pass": {
        "value": {},
        "description": "overide of the completion fraction of dataset with keyword"
    },
    "tiers_with_no_custodial": {
        "value": ["DQM", "DQMIO", "RECO", "RAWAODSIM"],
        "description": "The data tiers that do not go to tape. Can be overidden by custodial overide at campaign level"
    },
    "use_parent_custodial": {
        "value": False,
        "description": "Use the location of the parent dataset for custodial copy"
    },
    "tape_size_limit": {
        "value": 200,
        "description": "Size over which to prevent transfer to tape automatically"
    },
    "tiers_with_no_check": {
        "value": ["DQM", "DQMIO"],
        "description": "The data tiers that do not pass closeout checks. Can be overidden by custodial overide at campaign level"
    },
    "tiers_no_DDM": {
        "value": ["GEN-SIM", "LHE", "GEN", "DQM", "DQMIO", "GEN-SIM-DIGI-RAW", "RAW"],
        "description": "The data tiers that do not go to AnaOps"
    },
    "tiers_to_DDM": {
        "value": ["AODSIM", "MINIAODSIM", "GEN-SIM-RAW", "GEN-SIM-RECO", "GEN-SIM-RECODEBUG", "AOD", "RECO", "MINIAOD",
                  "ALCARECO", "USER", "RAW-RECO", "RAWAODSIM"],
        "description": "The data tiers that go to AnaOps"
    },
    "tiers_keep_on_disk": {
        "value": ["LHE"],
        "description": "the data tier not unlocked until used again"
    },
    "check_fullcopy_to_announce": {
        "value": False,
        "description": "Whether to check for a full copy being present prior to announcing a dataset"
    },
    "stagor_sends_back": {
        "value": True,
        "description": "Whether the stagor module can send workflow back to considered"
    },
    "max_handled_workflows": {
        "value": 4000,
        "description": "The total number of workflows that we allow to handle at a time (transfer, running, assistance)"
    },
    "max_staging_workflows": {
        "value": 700,
        "description": "The total number of workflows that we allow to stage at a time"
    },
    "max_staging_workflows_per_site": {
        "value": 700,
        "description": "The total number of workflows that we allow to stage at a time per site"
    },
    "max_transfer_in_GB": {
        "value": 800000,
        "description": "The total size of the input datasets that can be transfered at a given time"
    },
    "transfer_timeout": {
        "value": 7,
        "description": "Time in days after which to consider a transfer to be stuck"
    },
    "transfer_lowrate": {
        "value": 0.004,
        "description": "Rate in GB/s under which to consider a transfer to be stuck, after transfer_timeout days"
    },
    "less_copies_than_requested": {
        "value": 1,
        "description": "Decrease the number of requested copies by that number, floored to 1"
    },
    "chopping_threshold_in_GB": {
        "value": 4000,
        "description": "The threshold before choping an input dataset in chunk of that size for spreading to sites"
    },
    "error_codes_to_recover": {
        "value": {"50664": [{"legend": "time-out",
                             "solution": "split-2",
                             "details": None,
                             "rate": 20
                             }],
                  "50660": [{"legend": "memory excess",
                             "solution": "mem-1000",
                             "details": None,
                             "rate": 20
                             }],
                  "61104": [{"legend": "failed submit",
                             "solution": "recover",
                             "details": None,
                             "rate": 20
                             }],
                  "8028": [{"legend": "read error",
                            "solution": "recover",
                            "details": None,
                            "rate": 20
                            }],
                  "8021": [{"legend": "cmssw failure",
                            "solution": "recover",
                            "details": "FileReadError",
                            "rate": 20
                            }],
                  "71305": [{"legend": "long pending",
                             "solution": "recover",
                             "details": None,
                             "rate": 20
                             }],
                  "8001": [{"legend": "lhe failure",
                            "solution": "split-4",
                            "details": "No lhe event found in ExternalLHEProducer::produce()",
                            "rate": 20
                            }]
                  },
        "description": "The error code, threshold and rules for auto-recovery"
    },
    "error_codes_to_block": {
        "value":
            {
                "99109": [{"legend": "stage-out",
                           "solution": "recover",
                           "details": None,
                           "rate": 20
                           }]
            },
        "description": "The error code, threshold and rules to prevent auto-recovery"
    },
    "error_codes_to_notify": {
        "value": {
            "8021": {"message": "Please take a look and come back to Ops."}
        },
        "description": "The error code, threshold and rules to notify the user of an error in production"
    },
    "user_rereco": {
        "description": "The users from which we expect ReReco requests",
        "value": ["cerminar", "fabozzi"]
    },
    "user_relval": {
        "description": "The users from which we expect relval requests",
        "value": ["fabozzi", "nwickram", "bsutar", "rverma", "prebello", "piperov", "sandhya"]
    },
    "relval_routing": {
        "description": "Set of keywords and special settings for relvals",
        "value": {"cc7": {"parameters": {"SiteWhitelist": ["T2_US_Nebraska"]}},
                  "highIOtobedecidedflag": {"parameters": {"SiteWhitelist": ["T2_US_Nebraska", "T2_US_Purdue"]}}
                  }
    },
    "batch_goodness": {
        "description": "Level below which to include a note in the batch report",
        "value": 90
    }
}



#!/usr/bin/env python
"""
_Subscription.New_

Oracle implementation of Subscription.New
"""

from WMCore.WMBS.MySQL.Subscriptions.New import New as NewMySQL

class New(NewMySQL):
    typesSQL = """INSERT INTO wmbs_sub_types (name)
                    SELECT :subtype FROM dual
                    WHERE NOT EXISTS (SELECT id FROM wmbs_sub_types WHERE name = :subtype)"""

    sql = """INSERT INTO wmbs_subscription
             (fileset, workflow, subtype, split_algo, last_update)
             SELECT :fileset, :workflow, id, :split_algo, :timestamp
             FROM wmbs_sub_types WHERE name = :subtype"""



#!/usr/bin/env python

"""
Oracle implementation of AddCKType
"""





from WMCore.WMBS.MySQL.Files.AddCKType import AddCKType as MySQLAddCKType

class AddCKType(MySQLAddCKType):
    sql = """INSERT INTO wmbs_checksum_type (type)
                VALUES (:cktype)"""



"""
This module contains a few basic WMCore-related utilitarian functions
"""

from Utils.CertTools import ckey, cert

from WMCore.Services.pycurl_manager import RequestHandler


def makeHttpRequest(srvUrl, payload=None, method='GET', headers=None, encode=True, decode=True, payloadSizeToDump=50):
    """
    Perform HTTP call to provided service URL. This can be any of supported methods like
    GET, POST, PUT, DELETE. It will return HTTP response from upstream MicroService
    and parse it accordingly (the WMCore MS service returns results as {'result': {'data':...}}
    HTTP response.

    :param srvUrl: string with the service url
    :param payload: payload query or document, in case of GET HTTP request it is query dictionary,
        in case of POST/PUT HTTP requests it is JSON payload to the service
    :param method: string, defines which HTTP method to use
    :param headers: HTTP headers (presented as dictionarY)
    :param encode: boolean to reflect if data should be encoded, see pycurl_manager.py
    :param decode: boolean to reflect if data should be decoded, see pycurl_manager.py
    :param payloadSizeToDump: amount of characters to dump in a log from payload (to make log entries readable)
    :return: returns a list with all objects, or raises
        an exception in case of failure
    """
    payload = payload or {}
    mgr = RequestHandler()
    headers = headers or {'Content-Type': 'application/json'}
    data = mgr.getdata(srvUrl, payload, headers, verb=method,
                       ckey=ckey(), cert=cert(), encode=encode, decode=decode)
    if data and data.get("result", []):
        if "error" in data["result"][0]:
            # strip off part of payload to make readable log message
            sdata = f"{payload}"
            if len(sdata) > payloadSizeToDump:
                sdata = sdata[:payloadSizeToDump] + "..."
            msg = f"Failed to contact {srvUrl} via {method} request with {sdata}"
            msg += f" and error message: {data}"
            raise RuntimeError(msg)
    return data["result"]

def getPileupDocs(mspileupUrl, queryDict=None, method='GET'):
    """
    Fetch documents from MSPileup according to the query passed in using POST.

    :param mspileupUrl: string with the MSPileup url
    :param queryDict: dictionary with the MongoDB query to run
    :param method: string, defines which HTTP method to use
    :return: returns a list with all the pileup objects, or raises
        an exception in case of failure
    """
    return makeHttpRequest(mspileupUrl, payload=queryDict, method=method)



#!/usr/bin/env python
"""
CouchCollection.py

Created by Dave Evans on 2010-03-14.
Copyright (c) 2010 Fermilab. All rights reserved.
"""

from WMCore.ACDC.Collection import Collection
from WMCore.ACDC.CouchFileset import CouchFileset
from WMCore.Database.CouchUtils import connectToCouch


class CouchCollection(Collection):
    """
    Collection that can be stored in CouchDB.

    Required Args:
      database - CouchDB database instance name
      url - CouchDB Server URL
      name - name of the collection
    """

    def __init__(self, **options):
        Collection.__init__(self, **options)
        self.url = options.get("url")
        self.database = options.get("database")
        self.name = options.get("name")
        self.server = None
        self.couchdb = None

    @connectToCouch
    def drop(self):
        """
        _drop_

        Drop this collection and all files and filesets within it.
        """
        params = {"startkey": [self.name],
                  "endkey": [self.name, {}],
                  "reduce": False}
        result = self.couchdb.loadView("ACDC", "coll_fileset_docs",
                                       params)

        for row in result["rows"]:
            self.couchdb.delete_doc(row["id"])
        return

    @connectToCouch
    def populate(self):
        """
        _populate_

        The load the collection and all filesets and files out of couch.
        """
        params = {"startkey": [self.name],
                  "endkey": [self.name, {}],
                  "reduce": True, "group_level": 2}
        result = self.couchdb.loadView("ACDC", "coll_fileset_docs",
                                       params)
        self["filesets"] = []
        for row in result["rows"]:
            fileset = CouchFileset(database=self.database, url=self.url,
                                   name=row["key"][1])
            self.addFileset(fileset)
            fileset.populate()
        return



#!/usr/bin/env python

"""
Creates jobs for new subscriptions

"""

import logging
import threading

from WMComponent.JobSubmitter.JobSubmitterPoller import JobSubmitterPoller
from WMCore.Agent.Harness import Harness


class JobSubmitter(Harness):
    """
    Creates jobs for new subscriptions

    """

    def __init__(self, config):
        # call the base class
        Harness.__init__(self, config)
        self.pollTime = 1

        logging.info("JobSubmitter.__init__")

    def preInitialization(self):
        """
        Setup the worker thread for jobSubmitter

        """
        logging.info("JobSubmitter.preInitialization")

        # Add event loop to worker manager
        myThread = threading.currentThread()

        pollInterval = self.config.JobSubmitter.JobSubmitterPoller.pollInterval
        logging.info("Setting poll interval to %s seconds", pollInterval)
        myThread.workerThreadManager.addWorker(JobSubmitterPoller(self.config),
                                               pollInterval)

        return



#!/usr/bin/env python
"""
_LogCollect_

Diagnostic implementation for a job's LogCollect step


"""




from builtins import range
import os
from WMCore.WMSpec.Steps.Diagnostic import Diagnostic, DiagnosticHandler


class LCExceptionHandler(DiagnosticHandler):
    """
    _LCExceptionHandler_

    Generic handler for the LogCollect step

    I have no idea what this should do

    """
    def __call__(self, errCode, executor, **args):
        """
        _operator()_

        Twiddle thumbs, contemplate navel, toss coin

        """
        msg         = "Error in LogCollect: %s\n" % (errCode)
        description = "Misc. LogCollect error: %s\n" % (errCode)
        if args.get('ex', None):
            msg += str(args.get('ex'))
        if args.get('ExceptionInstance', False):
            msg += str(args.get('ExceptionInstance'))



        #jobRepXml = os.path.join(executor.step.builder.workingDir,
        #                         executor.step.output.jobReport)
        #
        #if not os.path.exists(jobRepXml):
        #    # no report => Error
        #    msg = "No Job Report Found: %s" % jobRepXml
        #    executor.report.addError(50115, "MissingJobReport", msg)
        #    return
        #
        ## job report XML exists, load the exception information from it
        #executor.report.parse(jobRepXml)


        # make sure the report has the error in it
        errSection = getattr(executor.report.report, "errors", None)
        if errSection == None:
            msg = "Job Report contains no error report, but LogCollect exited non-zero: %s" % errCode
            executor.report.addError(executor.stepName, 50116, "MissingErrorReport", msg)
            return

        else:
            #check exit code in report is non zero
            if executor.report.report.status == 0:
                msg = "Job Report contains no error report, but LogCollect exited non-zero: %s" % errCode
                executor.report.addError(executor.stepName, 50116, "MissingErrorReport", msg)

            executor.report.addError(executor.stepName,
                                    errCode, description, msg)
        return

class LogCollect(Diagnostic):

    def __init__(self):
        Diagnostic.__init__(self)


        catchAll = LCExceptionHandler()
        [ self.handlers.__setitem__(x, catchAll) for x in range(0, 255) if x not in self.handlers ]



#!/usr/bin/env python



"""
_RunJob_

The runJob class object.
It is very simple.
"""

from WMCore.WMBS.Job import Job


class RunJob(dict):
    """
    _RunJob_

    Basically, create an organized dictionary with all
    the necessary fields.
    """

    def __init__(self, jobid=-1):
        """
        Just make sure you init the dictionary fields.

        If the field has no value, leave it as None so we can
        overwrite it later.
        """

        self.setdefault('id', None)
        self.setdefault('jobid', jobid)
        self.setdefault('gridid', None)
        self.setdefault('bulkid', None)
        self.setdefault('retry_count', 0)
        self.setdefault('status', None)
        self.setdefault('location', None)
        self.setdefault('site_cms_name', None)
        self.setdefault('userdn', None)
        self.setdefault('usergroup', '')
        self.setdefault('userrole', '')
        self.setdefault('plugin', None)
        self.setdefault('cache_dir', None)
        self.setdefault('status_time', None)
        self.setdefault('packageDir', None)
        self.setdefault('sandbox', None)
        self.setdefault('wf_priority', None)
        self.setdefault('task_type', None)
        self.setdefault('possibleSites', None)
        self.setdefault('swVersion', None)
        self.setdefault('scramArch', None)
        self.setdefault('siteName', None)
        self.setdefault('name', None)
        self.setdefault('proxyPath', None)
        self.setdefault('request_name', None)
        self.setdefault('estimatedJobTime', None)
        self.setdefault('estimatedDiskUsage', None)
        self.setdefault('estimatedMemoryUsage', None)
        self.setdefault('numberOfCores', 1)
        self.setdefault('taskPriority', None)
        self.setdefault('task_name', None)
        self.setdefault('task_id', None)
        self.setdefault('potentialSites', None)
        self.setdefault('inputDataset', None)
        self.setdefault('inputDatasetLocations', None)
        self.setdefault('inputPileup', None)
        self.setdefault('allowOpportunistic', False)
        self.setdefault('activity', None)
        self.setdefault('requiresGPU', 'forbidden')
        self.setdefault('gpuRequirements', None)
        self.setdefault('jobExtraMatchRequirements', '')
        self.setdefault('campaignName', None)
        self.setdefault('requestType', None)
        self.setdefault('physicsTaskType', None)

        return

    def buildFromJob(self, job):
        """
        _buildFromJob_

        Build a RunJob from a WMBS Job
        """

        # These two are required
        self['jobid'] = job.get('id', None)
        self['retry_count'] = job.get('retry_count', None)
        self['userdn'] = job.get('owner', None)
        self['usergroup'] = job.get('usergroup', '')
        self['userrole'] = job.get('userrole', '')
        self['siteName'] = job.get('custom', {}).get('location', None)

        # Update the job with all other shared keys
        for key in job:
            if key in self:
                self[key] = job[key]

        return

    def buildWMBSJob(self):
        """
        _buildWMBSJob_

        Does exactly what it sounds like

        Also, attach couch_record (since we usually need one)
        """

        job = Job(id=self['jobid'])
        job['retry_count'] = self['retry_count']
        job['couch_record'] = None
        job['owner'] = self['userdn']
        job['usergroup'] = self['usergroup']
        job['userrole'] = self['userrole']

        for key in self:
            if key != 'id':
                job[key] = self[key]

        return job



import json

from WMCore.Services.Service import Service


class WMAgent(Service):

    """
    API for dealing with retrieving information from WMAgent RESTModel Service
    """

    def __init__(self, dict={}):
        """
        responseType will be JSON
        """

        dict.setdefault("accept_type", "application/json")
        dict.setdefault("content_type", "application/json")
        self.encoder = json.dumps
        self.decoder = json.loads

        Service.__init__(self, dict)

    def _getResult(self, callname, clearCache = True,
                   args = None, verb="GET", contentType = None):
        """
        _getResult_

        retrieve JSON/XML formatted information given the service name and the
        argument dictionaries

        TODO: Probably want to move this up into Service
        """
        result = ''
        # make base file name from call name.
        file = callname.replace("/", "_")
        if clearCache:
            self.clearCache(file, args, verb)

        # can't pass the decoder here since refreshCache wright to file
        f = self.refreshCache(file, callname, args, encoder = self.encoder,
                              verb = verb, contentType = contentType)
        result = f.read()
        f.close()
        result = self.decoder(result)

        return result

    def getAgentStatus(self, detail = False):
        """
        """
        callname = 'agentstatus'
        # TODO support detail flag if it is needed
        # need to convert to boolean on serverside
        #args = {'detail': detail}
        args = {}
        return self._getResult(callname, args = args)

    def getACDCInfo(self):
        """
        """
        callname = 'acdclink'
        return self._getResult(callname)



#!/usr/bin/env python
"""
_LogCollect_

Template for a LogCollect Step

Mostly borrowed from StageOut since they share a similar function

"""

from future.utils import viewitems


from WMCore.WMSpec.Steps.Template import Template
from WMCore.WMSpec.Steps.Template import CoreHelper


class LogCollectStepHelper(CoreHelper):
    """
    _StageOutStepHelper_

    Add API calls and helper methods to the basic WMStepHelper to specialise
    for StageOut tasks

    This is very similar to StageOut since they have essentially the same function

    """

    def disableRetries(self):
        """
            handy for testing, without the 10 minute retry loop
        """
        self.data.retryCount = 1
        self.data.retryDelay = 0

    def addOutputDestination(self, lfn):
        """
        Adds an out location to put a tarball of all logs
        """

    def cmsswSetup(self, cmsswVersion, **options):
        """
        _cmsswSetup_
        Provide setup details for CMSSW.
        cmsswVersion - required - version of CMSSW to use
        Optional:
        scramCommand - defaults to scramv1
        scramProject - defaults to CMSSW
        scramArch    - optional scram architecture, defaults to None
        buildArch    - optional scram build architecture, defaults to None
        softwareEnvironment - setup command to bootstrap scram,defaults to None
        """
        self.data.application.setup.cmsswVersion = cmsswVersion
        for k, v in viewitems(options):
            setattr(self.data.application.setup, k, v)
        return

    def getScramArch(self):
        """
        _getScramArch_

        Retrieve the scram architecture used for this step.
        """
        return self.data.application.setup.scramArch

    def getCMSSWVersion(self):
        """
        _getCMSSWVersion_

        Retrieve the version of the framework used for this step.
        """
        return self.data.application.setup.cmsswVersion

class LogCollect(Template):
    """
    _LogCollect_

    Tools for creating a template LogCollect Step

    """

    def install(self, step):
        step.stepType = "LogCollect"
        step.section_("logs")
        step.logcount = 0
        step.retryCount = 3
        step.retryDelay = 300
        step.application.section_("setup")
        step.application.setup.scramCommand = "scramv1"
        step.application.setup.scramProject = "CMSSW"
        step.application.setup.cmsswVersion = None
        step.application.setup.scramArch = None
        step.application.setup.buildArch = None
        step.application.setup.softwareEnvironment = None

    def helper(self, step):
        """
        _helper_

        Wrap the WMStep provided in the CMSSW helper class that
        includes the ability to add and manipulate the details
        of a CMSSW workflow step

        """
        return LogCollectStepHelper(step)



#!/usr/bin/env python2.5
"""
_Emulator_

Define the Interface for a Step Emulator


"""


from builtins import object
class Emulator(object):
    """
    _Emulator_

    Base class for an Emulator implementation for a given step

    """
    def initialise(self, executorInstance):
        """
        _initialise_

        Post ctor initialisation, shortcut to some of the standard variables provided by the executor

        """
        self.executor = executorInstance
        self.job = self.executor.job
        self.report = self.executor.report
        self.task = self.executor.task
        self.step = self.executor.step
        self.stepSpace = self.executor.stepSpace
        self.stepName = self.executor.stepName


    def pre(self):
        """
        _pre_

        Override pre step to emulate
        """
        return None


    def execute(self):
        """
        _emulate_

        Emulate the response to the step provided

        """
        msg = "WMStep.Steps.Emulator.emulate not implemented for "
        msg += "class %s" % self.__class__.__name__
        raise NotImplementedError(msg)

    def post(self):
        """
        _post_

        Override to emulate post execution step

        """
        return None



"""
A simple watchdog mechanism based on OS signal handling, implemented to help in the monitoring
of the health of all WMAgent components.

This AgentWatchdogPoller thread is to be watching the thread's of all other components
for run time overflow and will take care of proper restarts of the misbehaved ones.
While at the same time the AgentWatchdogScanner is to be endowed with a mechanism
to monitor the health of all other components' process trees and take the proper
(restart) actions in case of unhealthy component is found.


Here bellow follows the timeline of the interaction between a component's thread and a timer:

Timer duration = pollInt + runTimeEst*corrFactor + extraTimeout + random add on between 0-10%

          t --->

          ...---|--- One polling cycle -----|-One polling cycle-|------- One polling cycle ----------|---...
                |                           |                   |                                    |
Thread:   ...---o-------------------o-------o-----------o-------o----------------------------o-------o---...
                |--------var--------| const |----var----| const |-------------var------------| const |---...
                |       runtime     |pollInt|  runtime  |pollInt|      runtime               |pollInt|
                                    |                   |                                    |
                                    |(tResetEvent)      |                           ^        |(tRESTARTEvent)
                                    |                   |(tResetEvent)              |        |
Timer:          o------ timer ------|.........o         |                           |        o------ timer ----------------o
                                    V                   |                           |
                                    o------ timer ------|..........o                |(actionEvent)
                                                        V                           |
                                                        o--------- timer -----------o


And we shall call this component with the affectionate nickname: Cerberus ;)
"""

import threading
import os
import signal
import logging
import time
import subprocess
import psutil
import inspect
import random
import json
import re

from pprint import pformat

from Utils.ProcFS import processStatus
from Utils.Timers import timeFunction
from Utils.wmcoreDTools import getComponentThreads, _getConfigSubsections, restart, forkRestart, isComponentAlive, getThreadConfigSections
from WMComponent.AgentWatchdog.Timer import Timer, _countdown, WatchdogAction, TimerException
from WMCore.WorkerThreads.BaseWorkerThread import BaseWorkerThread
from WMCore.WMInit import connectToDB
from WMCore.Services.AlertManager.AlertManagerAPI import AlertManagerAPI


class AgentWatchdogPoller(BaseWorkerThread):
    """
    The basic AgentWatchdogPoller class
    """
    def  __init__(self, config):
        BaseWorkerThread.__init__(self)
        self.config = config
        self.mainThread = threading.currentThread()
        self.mainThread.setName('Cerberus')

        # Register the expected signal handler:
        self.expectedSignal = signal.SIGCONT
        signal.signal(self.expectedSignal, self.sigHandler)

        self.watchedPid = None
        self.watchdogTimeout = self.config.AgentWatchdog.watchdogTimeout
        self.pollInterval = self.config.AgentWatchdog.AgentWatchdogPoller.pollInterval
        self.watchedComponents = self.config.AgentWatchdog.watchedComponents
        self.runTimeCorrFactor = self.config.AgentWatchdog.runTimeCorrFactor
        self.actionLimit = self.config.AgentWatchdog.actionLimit
        self.timers = {}

        self.mainThread.parents = [ thread['pid'] for thread in processStatus(self.mainThread.native_id)]
        logging.info(f"Initialized with parents {self.mainThread.parents}.")
        self.mainThread.enum = [ pid.native_id for pid in threading.enumerate() ]
        logging.info(f"Initialized with the following threads: {self.mainThread.enum}.")
        self.mainThread.main = threading.main_thread().native_id
        logging.debug(f"Initialized with main_thread {self.mainThread.main}.")

        self.alertManagerUrl = self.config.Alert.alertManagerUrl
        self.alertDestinationMap = self.config.Alert.alertDestinationMap
        self.alertManager = AlertManagerAPI(self.alertManagerUrl)
        logging.info(f"Setting up an alertManager instance for AgentWatchdogPoller and redirecting alerts to: {self.alertManagerUrl}.")

    def sigHandler(self, sigNum, currFrame, **kwargs):
        """
        A simple call back function to be used for resetting timers in the AgentWatchdog class
        It will be actually redirecting the received signal to all of the currently running timers/threads,
        and only the one to whom this signal was intended would react, by recognizing the senders pid
        """
        currThread = threading.currentThread()
        logging.info(f"{currThread.name}, pid: {currThread.native_id}, Asynchronous sigHandler for sigNum: {sigNum}")
        return True

    def sendAlert(self, alertMessage, severity='low'):
        """
        A generic method for sending alerts from AgentWatchdogPoller
        :param alertMesasge: A string with the alert contents
        :param severity:     Default: 'low'
        NOTE: The rest of the alertAPI parameters come from the agent configuration at init time
        """
        currAgent = getattr(self.config.Agent, 'hostName')
        alertName = "AgentWatchdogPoller"
        summary = f"Alert from WMAgent {currAgent}"
        description = alertMessage
        service = f"AgentWatchdogPoller@{currAgent}"
        self.alertManager.sendAlert(alertName, severity, summary, description, service, tag=self.alertDestinationMap['alertAgentWatchdogPoller'])

    def alertAction(self, timerName):
        """
        A wrapper method to create an alert action for any of the WMAgent Watchdog timers
        :param timerName: The timer instance as linked at self.timers[<TimerName>]
        """
        timer = self.timers[timerName]

        # Call alert manager
        msg = f"alertAction from: {timer.name}: The component {timer.compName} has been running beyond its estimated runtime with a big margin."
        msg += f"This might be a signal for a badly performing component thread and might need attention."
        logging.warning(msg)
        self.sendAlert(msg)

    def restartUpdateAction(self, compName):
        """
        _restartUpdateAction_
        This is an action wrapper intended to act on both sides:
        * On the component side: executing forkRestart
        * On the AgentWatchdogPoller side: updating the respective timer's data
          reflecting the new component status
        :param *: Accepts all parameters valid for wmcoreDTolls.forkRestart
        """
        # DONE: * simplified  signature
        #       * reset action counter
        #       * call alert manager

        # Check the wrapped method signature
        # NOTE: If this is a wrapper accepting arbitrary set of arguments like:
        #       self.restartUpdateAction(self, *args, **kwArgs):
        #       when passed to the Timer class as a callback action, it will bypass the
        #       action signature check during the timer initialization. So it will need
        #       an the extra check here. Something like e.g.:
        #       try:
        #            actionSignature = inspect.signature(forkRestart)
        #            actionSignature.bind(*args, **kwArgs)
        #       except TypeError as ex:
        #            msg = f"The timer action wrapper method signature does not match the set of arguments provided. Error: {str(ex)}"
        #            raise TimerException(msg) from None
        #       The diff between those two checks (the one at the Timer class and this one)
        #       is only in the moment they are performed. This one here will be done
        #       during the timer's runtime when the action execution is attempted,
        #       while the former is performed during the timer's init time.

        # Execute fork Restart with the arguments provided to the wrapper
        forkRestart(config=self.config, componentsList=[compName])

        # Give the component some time to start:
        time.sleep(1)

        # Update the respective timer's data
        timer = self._findTimersByComp(compName)
        if timer:
            compPidTree = getComponentThreads(self.config, compName, quiet=True)
            if compPidTree:
                self.updateTimer(timer, compPidTree)

        # Call alert manager
        msg = f"A restartUpdateAction has been performed by {timer.name}."
        logging.warning(msg)
        self.sendAlert(msg)

    def _findTimersByPid(self, pid):
        """
        _findTimersByPid_
        """
        correctTimers = []
        for timer in self.timers.values():
            # logging.debug(f"timer: {timer}")
            # logging.debug(f"timer.expPids: {timer.expPids}")
            if pid in timer.expPids:
                correctTimers.append(timer)
                # NOTE: Once we move to one timer per thread (meaning many timers per compName),
                #       we should not break here, but rather return the full list of timers
                #       found to be associated with this component.
        return correctTimers

    def _findTimersByComp(self, compName):
        """
        _findTimersByComp
        """
        correctTimers = []
        for timer in self.timers.values():
            if timer.compName == compName:
                correctTimers.append(timer)
                # NOTE: Once we move to one timer per thread (meaning many timers per compName),
                #       we should not break here, but rather return the full list of timers
                #       found to be associated with this component.
        return correctTimers

    def restartTimer(self, timer, compPidTree):
        """
        _restartTimer_
        """
        expPids = compPidTree['RunningThreads']
        expPids.append(compPidTree['Parent'])
        expPids.append(self.mainThread.native_id)
        # expPids.extend([thr.native_id for thr in threading.enumerate()])
        expPids = list(set(expPids))
        timer.restart(expPids=expPids)

    def updateTimer(self, timer, compPidTree):
        """
        _updateTimer_
        """
        expPids = compPidTree['RunningThreads']
        expPids.append(compPidTree['Parent'])
        expPids.append(self.mainThread.native_id)
        # expPids.extend([thr.native_id for thr in threading.enumerate()])
        expPids = list(set(expPids))
        timer.update(expPids=expPids)

    def setupTimer(self, compName, threadConfigSection, timerName=None):
        """
        Crates a timer from a given thread configuration section and associates it with
        the respective component.
        :param compName:            The component name.
        :param threadConfigSection: The configuration section of the thread for which this timer is about to be created
        :param timerName:           The threadName for this timer. If missing, it will be named after
                                    the threadConfigSection itself. (Default: None)
        :return:                    A Watchdog.Timer instance
        """
        timerName = timerName or threadConfigSection.getInternalName()

        logging.info(f"Creating timer: {timerName}")

        # Here to walk the pidTree of the component and set all expected pids
        # which are to be allowed to reset the timer
        compPidTree = {}
        try:
            compPidTree = getComponentThreads(self.config, compName)
            logging.info(f"Current Process tree for: {compName}: {compPidTree}")
        except Exception as ex:
            logging.error(f"Exception was thrown while rebuilding the the process tree for component: {compName}")
            logging.error(f"The full Error was : {str(ex)}")

        if not compPidTree:
            logging.error(f"Could not rebuild the the process tree for component: {compName}")
            logging.error(f"Giving up on timer creation for component: {compName}")
            return None

        expPids = compPidTree['RunningThreads']
        expPids.append(compPidTree['Parent'])

        # Here to add the full set of possible origin pids due to the signal redirection
        # (current thread, main thread, parent thread, etc.)
        # NOTE: This will be needed in the case where all agent's components are not sending
        #       the reset signal directly to the timer's thread/process but rather to the
        #       main thread of AgentWatchdogPoller which is then to redirect this signal to the
        #       correct timer.
        expPids.append(self.mainThread.native_id)
        # expPids.extend([thr.native_id for thr in threading.enumerate()])
        expPids = list(set(expPids))

        # Here to find the correct timer's interval
        # NOTE: We estimate the timer's interval by:
        #       * Taking the initial runtime estimate from the configuration SubSection for the component's thread/process
        #       * Multiply it with the common correction factor defined for all threads at the AgentWatchodog configurartion
        #       * Merge it with the pollInterval from the respective config SubSection or if none is found there
        #         with the pollInterval for the pid with the longest polling cycle from the component's config Section itself
        #       * Merge it with the watchdog timeout, in order to implement some static
        #         hysteresis in the watchdog logic. In the future instead of the shortest polling cycle
        #         (which does not reflect how long a component has run, but rather for how long we wait
        #         between component runs), we should use an estimator based on the runtime distribution
        #         of the slowest thread in the component.
        #       * Add some random factor between 1-10% on top of it, in order to avoid periodical
        #         overlaps between the timer's interval and the component's polling cycle,
        #         which would cause oscillations (the component being periodically rebooted due
        #         to lost signals caused by the intervals overlaps explained above)

        # Find the pre-configured runtimeEstimate
        threadRunTimeEst = getattr(threadConfigSection, 'runTimeEst', None)
        # Give up. No runtime estimate can be found for this thread
        if not threadRunTimeEst:
            logging.error(f"Failed to estimate the pollInterval for the timer: {timerName}. This one will be skipped")
            return None

        # Apply the safety margin:
        threadRunTimeEst *= self.runTimeCorrFactor

        # Find the correct poll interval for this thread
        # TODO: To find the compConfigSection this threadConfigSection to
        threadPollInterval = getattr(threadConfigSection, 'pollInterval', None)
        if not threadPollInterval:
            # Take the longest interval:
            compPollIntervals = {attr: value
                                 for attr,value in inspect.getmembers(compConfigSection)
                                 if re.match(r"^.*[p,P]ollInterval$", attr)}
            try:
                maxPollIntervalName = max(compPollIntervals)
                threadPollInterval = compPollIntervals[maxPollIntervalName]
                logging.info(f"Selecting the longest polling interval in the component: {maxPollIntervalName}:{threadPollInterval} sec.")
            except (ValueError, KeyError) as ex:
                threadPollInterval = None
                logging.error(f"No pollIntervals defined for component: {compName}")

        # Give up after the final attempt to estimate the pooll interval
        if not threadPollInterval:
            logging.error(f"Failed to estimate the pollInterval for the timer: {timerName}. This one will be scipped")
            return None

        # Set the timerInterval as the sum of the previous two
        timerInterval = threadRunTimeEst + threadPollInterval

        # Now lets add some disturbance to the force:
        timerInterval += self.watchdogTimeout
        timerInterval *= random.uniform(1.01, 1.1)

        # Create the action to be executed at the end of the timer.
        # NOTE: The format is a named tuple of the form:
        #       (callbackFunction, [args], {kwArgs})
        # action = WatchdogAction(self.restartUpdateAction, [compName], {})
        action = WatchdogAction(self.alertAction, [timerName], {})
        logging.debug(f"action function: {action}.")

        # Here to define the timer's path, where it will be permanently written on disk
        compDir = self.config.section_(compName).componentDir
        compDir = os.path.expandvars(compDir)
        timerPath = f"{compDir}/Timer-{timerName}"

        try:
            timer = Timer(name=timerName,
                             compName=compName,
                             expPids=expPids,
                             action=action,
                             actionLimit=self.actionLimit,
                             path=timerPath,
                             interval=timerInterval)
            # Preserve the creator's PID inside the timer
            # NOTE: This is a MUST, because otherwise if the timer expires the component
            #       won't know to whom to send the signal for properly restarting it
            timer.creator_id = threading.current_thread().native_id

        except Exception as ex:
            logging.error(f"Failed to create timer: {timerName} for component {compName} due to ERROR: {str(ex)}")
            return None

        self.timers[timerName] = timer
        return timer

    def setupCompTimers(self, compName):
        """
        Creates a timer per thread/process for all component's threads/processes.
        We spawn a separate thread for every timer.
        :param compName: The component name.
        :return:         Nothing
        """
        currThread = threading.currentThread()
        logging.info(f"setting up timers for component: {compName}.")

        for threadName, threadConfigSection in getThreadConfigSections(self.config, compName).items():

            # Here to setup the individual timer and add it to the timers list in the AgentWatchdog object

            # Adding an exception only for WorkQueueManagerReqMgrPoller, since its activation at the component
            # depends on an additional config parameter marking if the queue is Local or Global:
            if threadName == 'WorkQueueManagerReqMgrPoller' and self.config.WorkQueueManager.level == 'LocalQueue':
                logging.warning(f"Skip creation of timer: {threadName} for a Local WorkQueue.")
                continue

            # NOTE: Currently we are assigning the timerName to the actual threadName instead of compName
            #       in contrast to what it was when we were creating a single timer per component
            if not self.setupTimer(compName, threadConfigSection, timerName=threadName):
                continue

            # Finally start the timers:
            self.timers[threadName].start()

            # Here to preserve the timer on disk, such that its parameters can later be found by the
            # components itself, and it can be reset on time.
            # NOTE: This must happen upon timer's startup, because otherwise the timer attributes
            #       would be incomplete and later the component won't be able to reset it.
            self.timers[threadName].write()

            # NOTE: The so preserved timer would reflect only the state of the timer
            #       at init time, and won't be updated later until recreated. So any
            #       dynamic property like endTime, remTime may not reflect the real
            #       state of the timer as it is currently in memory.
            # DONE: To implement a mechanism for rewriting the timer's content/state
            #       upon receiving a signal.SIGCONT. This way the main thread can update
            #       the contents of its timers on a regular basis(e.g. with a rate depending
            #       on the AgentWatchDog polling cycle).


    @timeFunction
    def algorithm(self, parameters=None):
        """
        Spawn and register watchdog timers
        """
        currThread = threading.currentThread()

        startTime = time.time()
        endTime = startTime + self.pollInterval

        # Check all components' health:
        # DONE: To move it in a separate thread, not to mess up with the blocking calls
        #       for refreshing the timers data on disk bellow. And here, mess up means delaying,
        #       because the calls to wmcoreD.isComponentAlive have non zero runtime)
        #       The example code bellow does not allow to rebuild the newly restarted
        #       components timer, since the timer lives in the main thread not in the child
        # #compAliveThread = threading.Thread(target=self.checkCompAlive, name="ComponentsWatcher")
        # #compAliveThread.start()
        # self.checkCompAlive()

        logging.debug(f"Polling cycle started with current pid: {currThread.native_id}.")
        logging.debug(f"Main pid from threading module: {threading.main_thread().native_id}, main pid at startup: {self.mainThread.native_id}, list of watched components: {list(self.timers.keys())}")
        logging.debug(f"Checking and Re-configuring previously expired timers for components with changed running state.")
        logging.debug(f"All current threads: {[thr.native_id for thr in  threading.enumerate()]}.")

        # Refresh timers:
        for compName in self.watchedComponents:
            timers = self._findTimersByComp(compName)

            # Retrying to create timers for components present in the watched list
            # but previously failed during initialization:
            logging.debug(f"Trying to recreate all previously failed timers for component: {compName}")
            configuredThreads = getThreadConfigSections(self.config, compName)
            for threadName, threadConfigSection in configuredThreads.items():
                # Adding the WorkQueueManagerReqMgrPoller exception for local WorkQueue yet again
                if threadName == 'WorkQueueManagerReqMgrPoller' and self.config.WorkQueueManager.level == 'LocalQueue':
                    continue
                if threadName not in [timer.name for timer in timers]:
                    if self.setupTimer(compName, threadConfigSection, timerName=threadName):
                        self.timers[threadName].start()
                        self.timers[threadName].write()

            # Re-configuring any timers associated with components which changed execution or have been restarted:
            for timer in timers:
                compPidTree = getComponentThreads(self.config, compName, quiet=True)
                if compPidTree and compPidTree['Parent'] not in timer.expPids:
                    if timer.is_alive():
                        logging.info(f"Re-configuring timer: {timer}, whose associated component has changed state or have been restarted.")
                        self.updateTimer(timer, compPidTree)
                    else:
                        logging.info(f"Restarting expired timer: {timer}, whose associated component has changed state or have been restarted.")
                        self.restartTimer(timer, compPidTree)

        # Refresh all timers' data on disk every 1 second. And restart expired timers
        # if a signal at the main thread has been received.
        # NOTE: When a signal.SIGCONT is received at the AgentWatchdogPooller's thread from a
        #       particular component's thread it means an overflowing thread has tried to resets its timer
        #       but failed because the timer has already expired. In such case we should consider the thread
        #       good to be watched again, since it did manage to complete its polling cycle and has made
        #       an attempt to reset the timer. In this situation we are about to find the timer per PID
        #       and just restart it.
        while _countdown(endTime):
            sigInfo = signal.sigtimedwait([self.expectedSignal], 1)
            if sigInfo:
                # Restarting a particular timer depending on the sender's PID:
                logging.info(f"{self.mainThread.name} with main pid: {threading.main_thread().native_id}: Received signal: {pformat(sigInfo)}")

                # First, find the correct timers:
                # NOTE: The condition are:
                #       * The sender's PID has to belong to a recognizable component's timer - This check
                #         would return more than a single timer for a multi-threaded component
                #       * Out of all such found timers filter only those which has already expired
                # TODO: To narrow the set of expPids per component such that we can have one and only one
                #       PID which should be allowed to reset the timer - this would avoid false positives.
                #       But this would require to implement a call to `pthread_setname_np` method from the `pthread`
                #       C library through ctype in order to preserve the thread's name at the OS in the proc fiel system.
                #       and later fetched through `psutils`
                timers = [timer for timer in self._findTimersByPid(sigInfo.si_pid) if not timer.alive]

                if not timers:
                    # Ignore signals from unknown origin:
                    logging.warning(f"The sender's pid: {sigInfo.si_pid} was not recognized or NO expired timers could be associate with it.")
                    logging.warning(f"Ignoring the signal.")
                    continue

                # Second, Refresh timer data:
                for timer in timers:
                    try:
                        logging.info(f"Restarting timer: {timer.name}")
                        self.restartTimer(timer, getComponentThreads(self.config, timer.compName, quiet=True))
                    except Exception:
                        logging.warning(f"Could not restart {timer.name}. The current signal is lost.")

            else:
                logging.debug(f"Refreshing all timers' data on disk.")
                for timer in self.timers.values():
                    timer.write()

        logging.debug(f"Reached the end of its polling cycle.")


        # NOTE: This whole logic bellow is for the case where we use signal redirection from the main AgentWatchdog thread
        #       to the timers. Currently we already took the path for communicating the timers' data through files
        #       so this mechanism is no longer needed and commented out for historical purposes. The logic
        #       is refurbished and reused (in the above paragraph) for the mechanism where the main thread is about to refresh all timers'
        #       data on disk upon receiving a signal.SIGCONT (instead of finding the correct timer and resetting it)

        # while True:
        #     sigInfo = signal.sigtimedwait([self.expectedSignal], _countdown(endTime))
        #     if sigInfo:
        #         # Resetting the correct timer:
        #         logging.info(f"{self.mainThread.name} with main pid: {threading.main_thread().native_id}: Received signal: {pformat(sigInfo)}")

        #         # First, find the correct timer:
        #         correctTimer = None
        #         for timer in self.timers.values():
        #             # logging.debug(f"timer: {timer}")
        #             # logging.debug(f"timer.expPids: {timer.expPids}")
        #             if sigInfo.si_pid in timer.expPids:
        #                 correctTimer = timer
        #                 break

        #         # Second, redirect the signal to the timer's thread:
        #         if correctTimer:
        #             # try to reset it:
        #             try:
        #                 logging.info(f"Resetting timer: {correctTimer.name}")
        #                 logging.info(f"Timer expPids: {correctTimer.expPids}")
        #                 logging.info(f"sending signal from: {currThread.native_id}")
        #                 logging.info(f"sending signal to: {correctTimer.native_id}")
        #                 os.kill(correctTimer.native_id, self.expectedSignal)
        #             except ProcessLookupError:
        #                 logging.warning(f"Missing timer: {correctTimer.name}. It will be recreated on the next AgentWatchdogPoller cycle, but the current signal is lost and the component may be restarted soon.")
        #         else:
        #             # Ignore signals from unknown origin:
        #             logging.info(f"The sender's pid: {sigInfo.si_pid} was not recognized. Ignoring the signal")
        #             continue
        #     else:
        #         logging.info(f"Reached the end of polling cycle.")
        #         break


    def setup(self, parameters=None):
        """
        __setup__

        Setup the name of the component to be watched
        """
        # Wait one full cycle before starting the whole component to allow all others to initialize properly
        logging.info("Waiting for one cycle to let others initialize properly ... ")
        # time.sleep(self.pollInterval)
        time.sleep(10)

        logging.info("Setting up all timers ... ")

        for compName in self.watchedComponents:
            self.setWatchedComponent(compName)
        return

    def setWatchedPid(self, pid):
        """
        A simple function to add the watched component PID
        :param pid: The PID to be added
        """
        self.watchedPid = pid

    def setWatchedComponent(self, compName=None):
        """
        A simple function to add a watched component. We start a single timer per component
        And the component's thread with the shortest polling interval is supposed to reset the timer.
        :param compName: the component name to be watched (Default: AgentStatusWatcher')
        """
        logging.info(f"Start watching {compName}")
        self.setupCompTimers(compName)



# pylint: disable-msg=C0103,C0111,W0212
"""
Throttled module defines necessary classes to throttle context.
Should be used in Web server APIs which wants to throttle clients.

The counter based approach, implemented in _ThrottleCounter class,
defines throttling on total number of user access to the underlying
app. But its counter depends on time execution of the API.

The time based approach, implemented in _ThrottleTimeCount class,
count total number of user accesses within specified time interval
regardless of time execution of the underlying API(s).

Here is an example how to use this module/classes:

# counter based throttline

from Utils.Throttling import UserThrottle
thr = UserThrottle(limit=5) # adjust threshold limit here
@thr.make_throttled()
def api():
    # define your api logic here

# time based throttling

from Utils.Throttling import UserThrottleTime
thr = UserThrottleTime(limit=5) # adjust threshold limit here
@thr.make_throttled(trange=2) # adjust trange (in sec) here
def api():
    # define your api logic here

"""


# standard modules
import time
from builtins import object
import logging
import threading

# cherrypy modules
import cherrypy


class _ThrottleCounter(object):
    """
    _ThrottleCounter class defines throttle parameter and
    enter/exit methods to work with `with` context.
    """

    def __init__(self, throttle, user, debug=False):
        self.throttle = throttle
        self.user = user
        if debug:
            self.throttle.logger.setLevel(logging.DEBUG)

    def __enter__(self):
        "Define enter method for `with` context"
        ctr = self.throttle._incUser(self.user)
        msg = "Entering throttled function with counter %d for user %s" \
                % (ctr, self.user)
        self.throttle.logger.debug(msg)
        if ctr > self.throttle.getLimit():
            msg = "The current number of active operations for this resource"
            msg += " exceeds the limit of %d for user %s" \
                   % (self.throttle.getLimit(), self.user)
            raise cherrypy.HTTPError("429 Too Many Requests", msg)

    def __exit__(self, exc_type, exc_value, exc_traceback):
        "Define exit method for `with` context"
        ctr = self.throttle._decUser(self.user)
        msg = "Exiting throttled function with counter %d for user %s" \
                % (ctr, self.user)
        self.throttle.logger.debug(msg)

class _ThrottleTimeCounter(object):
    """
    _ThrottleTimeCounter class defines time range throttled mechanism
    """

    def __init__(self, throttle, user, trange):
        self.throttle = throttle
        self.user = user
        self.trange = trange

    def __enter__(self):
        "Define enter method for `with` context"
        ctr = self.throttle._incUser(self.user, self.trange)
        if ctr > self.throttle.getLimit():
            self.throttle._decUser(self.user)
            msg = "The current number of active operations for this resource"
            msg += " exceeds the limit of %d for user %s in last %s sec" \
                   % (self.throttle.getLimit(), self.user, self.trange)
            raise cherrypy.HTTPError("429 Too Many Requests", msg)

    def __exit__(self, exc_type, exc_value, exc_traceback):
        "Define exit method for `with` context"
        pass


class UserThrottleTime(object):
    """
    UserThrottle class defines how to handle throttle time range based mechanism.
    """

    def __init__(self, limit=3):
        self.lock = threading.Lock()
        self.users = {}
        self.users_time = {}
        self.limit = limit

    def getLimit(self):
        "Return throttle limit"
        return self.limit

    def throttleContext(self, user, trange=60):
        """
        Define throttle context via _ThrottleTimeCounter class
        which will count number of requests given user made in trange
        :param user: user name
        :param trange: time range while keep user access history
        """
        return _ThrottleTimeCounter(self, user, trange)

    def make_throttled(self, trange=60):
        "decorator for throttled context"
        def throttled_decorator(fn):
            def throttled_wrapped_function(*args, **kw):
                username = cherrypy.request.user.get('login', 'Unknown') \
                        if hasattr(cherrypy.request, 'user') else 'Unknown'
                with self.throttleContext(username, trange):
                    return fn(*args, **kw)
            return throttled_wrapped_function
        return throttled_decorator

    def reset(self, user):
        "Reset user activities, i.e. access counter and last time of access"
        self.users[user] = 0
        self.users_time[user] = time.time()

    def _incUser(self, user, trange=60):
        "increment user count"
        with self.lock:
            self.users.setdefault(user, 0)
            last_time = self.users_time.setdefault(user, time.time())
            # increase counter within our trange
            if abs(time.time()-last_time) < trange:
                self.users[user] += 1
            else:
                self.reset(user)
        return self.users[user]

    def _decUser(self, user, trange=60):
        "decrecrement user count"
        with self.lock:
            # decrease counter outside of our trange
            last_time = self.users_time[user]
            if abs(time.time()-last_time) < trange:
                # gradually decrease user counter based on his/her elapsed time
                step = int(trange - abs(time.time()-last_time))
                self.users[user] -= step
                if self.users[user] < 0:
                    self.reset(user)
            else:
                self.reset(user)

class UserThrottle(object):
    """
    UserThrottle class defines throttle counter based mechanism.
    """

    def __init__(self, limit=3):
        self.lock = threading.Lock()
        self.tls = threading.local()
        self.users = {}
        self.limit = limit
        self.logger = logging.getLogger("WMCore.UserThrottle")

    def getLimit(self):
        "Return throttle limit"
        return self.limit

    def throttleContext(self, user, debug=False):
        "defint throttle context"
        self.users.setdefault(user, 0)
        return _ThrottleCounter(self, user, debug)

    def make_throttled(self, debug=False):
        """
        decorator for throttled context
        """
        def throttled_decorator(fn):
            """
            A decorator
            """
            def throttled_wrapped_function(*args, **kw):
                """
                A wrapped function.
                """
                username = cherrypy.request.user.get('login', 'Unknown') \
                        if hasattr(cherrypy.request, 'user') else 'Unknown'
                with self.throttleContext(username, debug):
                    return fn(*args, **kw)
            return throttled_wrapped_function
        return throttled_decorator

    def _incUser(self, user):
        "increment user count"
        retval = 0
        with self.lock:
            retval = self.users[user]
            if getattr(self.tls, 'count', None) is None:
                self.tls.count = 0
            self.tls.count += 1
            if self.tls.count == 1:
                self.users[user] = retval + 1
        return retval

    def _decUser(self, user):
        "decrecrement user count"
        retval = 0
        with self.lock:
            retval = self.users[user]
            self.tls.count -= 1
            if self.tls.count == 0:
                self.users[user] = retval - 1
        return retval



#!/usr/bin/env python
"""
_IncrementRetry_

"""




from WMCore.Database.DBFormatter import DBFormatter

class IncrementRetry(DBFormatter):
    select = "SELECT id, retry_count FROM wmbs_job WHERE id = :job"
    update = "UPDATE wmbs_job SET retry_count = :retry WHERE id = :job"

    def execute(self, jobs = [], increment = 1, conn = None,
                transaction = False):
        binds = []
        for job in jobs:
            binds.append({"job": job["id"]})

        results = self.dbi.processData(self.select, binds, conn = conn,
                                       transaction = transaction)

        updateBinds = []
        for result in self.formatDict(results):
            updateBinds.append({"job": result["id"],
                                "retry": int(result["retry_count"]) + increment})

        self.dbi.processData(self.update, updateBinds, conn = conn,
                             transaction = transaction)
        return



"""
_RunJobByStatus_

Oracle implementation of Monitoring.RunJobByStatus
"""
from __future__ import print_function, division

from WMCore.BossAir.MySQL.RunJobByStatus import RunJobByStatus as MySQLRunJobByStatus

class RunJobByStatus(MySQLRunJobByStatus):
    pass



import logging
import random
import time
from json import JSONEncoder

import WMCore
from Utils.IteratorTools import grouper
from WMCore.Database.CMSCouch import CouchNotFoundError
from WMCore.Services.WMStats.WMStatsReader import WMStatsReader


def monitorDocFromRequestSchema(schema):
    """
    prun and convert
    """
    doc = {}
    doc["_id"] = schema['RequestName']
    doc["workflow"] = schema['RequestName']
    doc["requestor"] = schema['Requestor']
    doc["campaign"] = schema['Campaign']
    doc["request_type"] = schema['RequestType']
    doc["priority"] = schema['RequestPriority']
    doc["group"] = schema['Group']
    doc["request_date"] = schema['RequestDate']
    doc["type"] = "reqmgr_request"
    # additional field
    doc["inputdataset"] = schema.get('InputDataset', "")
    # additional field for Analysis work
    doc["vo_group"] = schema.get('VoGroup', "")
    doc["vo_role"] = schema.get('VoRole', "")
    doc["user_dn"] = schema.get('RequestorDN', "")
    doc["async_dest"] = schema.get('asyncDest', "")
    doc["dbs_url"] = schema.get("DbsUrl", "")
    doc["publish_dbs_url"] = schema.get("PublishDbsUrl", "")
    doc["outputdatasets"] = schema.get('OutputDatasets', [])
    doc["cmssw"] = schema.get('CMSSWVersion', [])
    doc['prep_id'] = schema.get('PrepID', None)

    # team name is not yet available need to be updated in assign status
    # doc['team'] = schema['team']
    return doc


def convertToServiceCouchDoc(wqInfo, wqURL):
    """
    Convert services generic info into a proper couch doc.
    """
    wqDoc = {}
    wqDoc['_id'] = wqURL
    wqDoc['agent_url'] = wqURL
    wqDoc['agent_team'] = ""
    wqDoc['agent_version'] = WMCore.__version__
    wqDoc['timestamp'] = int(time.time())
    wqDoc['down_components'] = []
    wqDoc['type'] = "agent_info"
    wqDoc.update(wqInfo)

    return wqDoc


class WMStatsWriter(WMStatsReader):
    def __init__(self, couchURL, appName="WMStats", reqdbURL=None, reqdbCouchApp="ReqMgr"):
        # set the connection for local couchDB call
        WMStatsReader.__init__(self, couchURL, appName, reqdbURL, reqdbCouchApp)

    def _sanitizeURL(self, couchURL):
        """
        don't sanitize url for writer
        """
        return couchURL

    def uploadData(self, docs):
        """
        upload to given couchURL using cert and key authentication and authorization
        """
        # add delete docs as well for the compaction
        # need to check whether delete and update is successful
        if isinstance(docs, dict):
            docs = [docs]
        for doc in docs:
            self.couchDB.queue(doc)
        return self.couchDB.commit(returndocs=True)

    def bulkUpdateData(self, docs, existingDocs):
        """
        Update documents to WMStats in bulk, breaking down to 100 docs chunks.
        :param docs: docs to insert or update
        :param existingDocs: dict of docId: docRev of docs already existent in wmstats
        """
        if isinstance(docs, dict):
            docs = [docs]
        for chunk in grouper(docs, 100):
            for doc in chunk:
                if doc['_id'] in existingDocs:
                    revList = existingDocs[doc['_id']].split('-')
                    # update the revision number and keep the history of the revision
                    doc['_revisions'] = {"start": int(revList[0]) + 1, "ids": [str(int(revList[1]) + 1), revList[1]]}
                else:
                    # then create a random 10 digits uid for the first revision number, required by new_edits=False
                    firstId = "%10d" % random.randrange(9999999999)
                    doc['_revisions'] = {"start": 1, "ids": [firstId]}
                self.couchDB.queue(doc)

            logging.info("Committing bulk of %i docs ...", len(chunk))
            self.couchDB.commit(new_edits=False)
        return

    def insertRequest(self, schema):
        doc = monitorDocFromRequestSchema(schema)
        return self.insertGenericRequest(doc)

    def updateTeam(self, request, team):
        return self.couchDB.updateDocument(request, self.couchapp, 'team',
                                           fields={'team': team})

    def insertTotalStats(self, request, totalStats):
        """
        update the total stats of given workflow (total_jobs, input_events, input_lumis, input_num_files)
        """
        return self.couchDB.updateDocument(request, self.couchapp, 'totalStats',
                                           fields=totalStats)

    def updateFromWMSpec(self, spec):
        # currently only update priority and siteWhitelist and output dataset
        # complex field needs to be JSON encoded
        # assuming all the toplevel tasks has the same site white lists
        # priority is priority + user priority + group priority
        fields = {'priority': spec.priority(),
                  'site_white_list': spec.getTopLevelTask()[0].siteWhitelist(),
                  'outputdatasets': spec.listOutputDatasets()}
        return self.couchDB.updateDocument(spec.name(), self.couchapp,
                                           'generalFields',
                                           fields={'general_fields': JSONEncoder().encode(fields)})

    def updateRequestsInfo(self, docs):
        """
        bulk update for request documents.
        TODO: change to bulk update handler when it gets supported
        """
        for doc in docs:
            del doc['type']
            self.couchDB.updateDocument(doc['workflow'], self.couchapp,
                                        'generalFields',
                                        fields={'general_fields': JSONEncoder().encode(doc)})

    def updateAgentInfo(self, agentInfo, propertiesToKeep=None):
        """
        replace the agentInfo document with new one.
        :param agentInfo: dictionary for agent info
        :param propertiesToKeep: list of properties to keep original value
        :return: None
        """
        try:
            exist_doc = self.couchDB.document(agentInfo["_id"])
            agentInfo["_rev"] = exist_doc["_rev"]
            if propertiesToKeep and isinstance(propertiesToKeep, list):
                for prop in propertiesToKeep:
                    if prop in exist_doc:
                        agentInfo[prop] = exist_doc[prop]

        except CouchNotFoundError:
            # this means document is not exist so we will just insert
            pass
        finally:
            result = self.couchDB.commitOne(agentInfo)
        return result

    def updateAgentInfoInPlace(self, agentURL, agentInfo):
        """
        :param agentInfo: dictionary for agent info
        :return: document update status

        update agentInfo in couch in place without replacing a doucment
        """
        return self.couchDB.updateDocument(agentURL, self.couchapp, 'agentInfo', fields=agentInfo)

    def updateLogArchiveLFN(self, jobNames, logArchiveLFN):
        for jobName in jobNames:
            self.couchDB.updateDocument(jobName, self.couchapp,
                                        'jobLogArchiveLocation',
                                        fields={'logArchiveLFN': logArchiveLFN})

    def deleteOldDocs(self, days):
        """
        delete the documents from wmstats db older than param 'days'
        """
        sec = int(days * 24 * 60 * 60)
        threshold = int(time.time()) - sec
        options = {"startkey": threshold, "descending": True,
                   "stale": "update_after"}
        result = self._getCouchView("time", options)

        for row in result['rows']:
            doc = {}
            doc['_id'] = row['value']['id']
            doc['_rev'] = row['value']['rev']
            self.couchDB.queueDelete(doc)
        committed = self.couchDB.commit()

        if committed:
            errorReport = {}
            deleted = 0
            for data in committed:
                if 'error' in data:
                    errorReport.setdefault(data['error'], 0)
                    errorReport[data['error']] += 1
                else:
                    deleted += 1
            return {'delete': deleted, 'error': errorReport}
        else:
            return "nothing"

    def replicate(self, target):
        return self.couchServer.replicate(self.dbName, target, continuous=True,
                                          filter='WMStats/repfilter')

    def getActiveTasks(self):
        """
        This is in Writter instance since it needs admin permission
        """
        couchStatus = self.couchServer.status()
        return couchStatus['active_tasks']

    def deleteDocsByIDs(self, ids):

        if not ids:
            # if ids is empty don't run this
            # it will delete all the docs
            return None

        docs = self.couchDB.allDocs(keys=ids)['rows']
        for j in docs:
            doc = {}
            doc["_id"] = j['id']
            doc["_rev"] = j['value']['rev']
            self.couchDB.queueDelete(doc)
        committed = self.couchDB.commit()
        return committed

    def replaceRequestTransitionFromReqMgr(self, docs):
        """
        bulk update for request documents.
        TODO: change to bulk update handler when it gets supported
        """

        for doc in docs:
            requestName = doc["RequestName"]
            requestTransition = {}
            requestTransition['request_status'] = []
            for r in doc["RequestTransition"]:
                newR = {}
                newR['status'] = r['Status']
                newR['update_time'] = r['UpdateTime']
                requestTransition['request_status'].append(newR)

            self.couchDB.updateDocument(requestName, self.couchapp,
                                        'generalFields',
                                        fields={'general_fields': JSONEncoder().encode(requestTransition)})

    def deleteDocsByWorkflow(self, requestName):
        """
        delete all wmstats docs for a given requestName
        """
        options = {"reduce": False, "key": requestName}
        docs = self._getCouchView("allWorkflows", options)['rows']

        for j in docs:
            doc = {}
            doc["_id"] = j['value']['id']
            doc["_rev"] = j['value']['rev']
            self.couchDB.queueDelete(doc)
        committed = self.couchDB.commit()
        return committed



#!/usr/bin/env python
"""
_RucioInjectorPoller_

General reminder about PhEDEx and Rucio definitions:
* CMS dataset is equivalent to a Rucio container
* CMS block is equivalent to a Rucio dataset
* CMS file is equivalent to a Rucio file
* A Rucio replica is a file, under a given scope, at a given RSE
"""

import json
import logging
import threading
import time
from pprint import pformat

from Utils.MemoryCache import MemoryCache
from Utils.Timers import timeFunction
from Utils.wmcoreDTools import resetWatchdogTimer, moduleName
from WMCore.DAOFactory import DAOFactory
from WMCore.Services.Rucio.Rucio import Rucio, WMRucioException
from WMCore.Services.Rucio.RucioUtils import RUCIO_VALID_PROJECT, RUCIO_RULES_PRIORITY
from WMCore.WMException import WMException
from WMCore.WorkerThreads.BaseWorkerThread import BaseWorkerThread
from Utils.IteratorTools import grouper


class RucioInjectorException(WMException):
    """
    _RucioInjectorException_

    Specific RucioInjectorPoller exception handling.
    """


class RucioInjectorPoller(BaseWorkerThread):
    """
    _RucioInjectorPoller_

    Poll the DBSBuffer database and inject files as they are created.

    The logic of this component is:
      * create a rucio container (or reuse a pre-existent one)
      * create a CMS block (or reuse a pre-existent one), block gets automatically attached
      * create file/replicas, which get automatically attached to its block as well
      * now create a CMS block rule to protect this data
      * if the block has been inserted into DBS, close the block in Rucio

    In addition to that, it has logic for rucio container subscription (rule creation),
    and block rule removal. Those follow a different polling cycle though.
    """

    def __init__(self, config):
        """
        ___init___

        Initialise class members
        """
        BaseWorkerThread.__init__(self)

        # dataset rule creation has a larger polling cycle
        self.config = config
        self.pollRules = config.RucioInjector.pollIntervalRules
        self.lastRulesExecTime = 0
        self.createBlockRules = config.RucioInjector.createBlockRules
        self.containerDiskRuleParams = config.RucioInjector.containerDiskRuleParams
        self.blockRuleParams = config.RucioInjector.blockRuleParams
        self.containerDiskRuleRSEExpr = config.RucioInjector.containerDiskRuleRSEExpr
        if config.RucioInjector.metaDIDProject not in RUCIO_VALID_PROJECT:
            msg = "Component configured with an invalid 'project' DID: %s"
            raise RucioInjectorException(msg % config.RucioInjector.metaDIDProject)
        self.metaDIDProject = dict(project=config.RucioInjector.metaDIDProject)

        # setup cache for container and blocks (containers can be much longer, make 6 days now)
        self.containersCache = MemoryCache(config.RucioInjector.cacheExpiration * 3, set())
        self.blocksCache = MemoryCache(config.RucioInjector.cacheExpiration, set())

        self.scope = getattr(config.RucioInjector, "scope", "cms")
        self.rucioAcct = config.RucioInjector.rucioAccount
        self.rucio = Rucio(acct=self.rucioAcct,
                           hostUrl=config.RucioInjector.rucioUrl,
                           authUrl=config.RucioInjector.rucioAuthUrl,
                           configDict={'logger': self.logger})

        self.useDsetReplicaDeep = getattr(config.RucioInjector, "useDsetReplicaDeep", False)
        self.delBlockSlicesize = getattr(config.RucioInjector, "delBlockSlicesize", 100)
        self.blockDeletionDelayHours = getattr(config.RucioInjector, "blockDeletionDelayHours", 0)
        self.blockDeletionDelaySeconds = self.blockDeletionDelayHours * 3600

        # metadata dictionary information to be added to block/container rules
        # cannot be a python dictionary, but a JSON string instead
        self.metaData = json.dumps(dict(agentHost=config.Agent.hostName,
                                        userAgent=config.Agent.agentName))

        self.testRSEs = config.RucioInjector.RSEPostfix
        self.filesToRecover = []

        # output data placement has a different behaviour between T0 and Production agents
        if hasattr(config, "Tier0Feeder"):
            logging.info("RucioInjector running on a T0 WMAgent")
            self.isT0agent = True
        else:
            self.isT0agent = False

        # NOTE: Setting to None all attributes inside __init__ to suppress pylint warnings
        self.getUninjected = None
        self.getMigrated = None
        self.getUnsubscribedBlocks = None
        self.setBlockRules = None
        self.findDeletableBlocks = None
        self.getCompletedBlocks = None
        self.markBlocksDeleted = None
        self.getUnsubscribedDsets = None
        self.markSubscribed = None
        self.setStatus = None
        self.setBlockClosed = None
        self.getDeletableWorkflows = None

        logging.info("Component configured to create block rules: %s", self.createBlockRules)

    def setup(self, parameters):
        """
        _setup_

        Create DAO Factory and setup some DAO.
        """
        myThread = threading.currentThread()
        daofactory = DAOFactory(package="WMComponent.RucioInjector.Database",
                                logger=self.logger, dbinterface=myThread.dbi)

        self.getUninjected = daofactory(classname="GetUninjectedFiles")
        self.getMigrated = daofactory(classname="GetMigratedBlocks")

        self.getUnsubscribedBlocks = daofactory(classname="GetUnsubscribedBlocks")
        self.setBlockRules = daofactory(classname="SetBlocksRule")

        self.findDeletableBlocks = daofactory(classname="GetDeletableBlocks")
        self.getCompletedBlocks = daofactory(classname="GetCompletedBlocks")
        self.markBlocksDeleted = daofactory(classname="MarkBlocksDeleted")
        self.getUnsubscribedDsets = daofactory(classname="GetUnsubscribedDatasets")
        self.markSubscribed = daofactory(classname="MarkDatasetSubscribed")

        daofactory = DAOFactory(package="WMComponent.DBS3Buffer",
                                logger=self.logger, dbinterface=myThread.dbi)
        self.setStatus = daofactory(classname="DBSBufferFiles.SetPhEDExStatus")
        self.setBlockClosed = daofactory(classname="SetBlockClosed")

        daofactory = DAOFactory(package="WMCore.WMBS",
                                logger=self.logger, dbinterface=myThread.dbi)
        self.getDeletableWorkflows = daofactory(classname="Workflow.GetDeletableWorkflows")

    @timeFunction
    def algorithm(self, parameters):
        """
        _algorithm_

        Poll the database for uninjected files and inject them into Rucio.
        """
        logging.info("Running Rucio injector poller algorithm...")

        try:
            # files that failed to get their status updated in dbsbuffer
            self._updateLFNState(self.filesToRecover, recovery=True)

            # get dbsbuffer_file.in_phedex = 0
            uninjectedFiles = self.getUninjected.execute()

            # create containers in rucio  (and update local cache)
            containersAdded = self.insertContainers(uninjectedFiles)
            if self.containersCache.isCacheExpired():
                self.containersCache.setCache(containersAdded)
            else:
                self.containersCache.addItemToCache(containersAdded)

            # create blocks. Only update the cache once a rule gets created...
            blocksAdded = self.insertBlocks(uninjectedFiles)
            if self.blocksCache.isCacheExpired():
                self.blocksCache.setCache(blocksAdded)
            else:
                self.blocksCache.addItemToCache(blocksAdded)

            # create file replicas
            self.insertReplicas(uninjectedFiles)

            # now close blocks already uploaded to DBS
            self.closeBlocks()

            if self.lastRulesExecTime + self.pollRules <= int(time.time()):
                self.insertContainerRules()
                self.insertBlockRules()
                self.deleteBlocks()
        except Exception as ex:
            msg = "Caught unexpected exception in RucioInjector. Details:\n%s" % str(ex)
            logging.exception(msg)
            raise RucioInjectorException(msg) from None

        # Reset its own watchdog timer at the end of the run cycle
        logging.info(f"Resetting {moduleName(self)} watchdog timer.")
        if resetWatchdogTimer(self):
            logging.warning(f"Failed to reset {moduleName(self)} watchdog timer. The component might be restarted soon.")

        return

    def insertContainers(self, uninjectedData):
        """
        This method will insert containers into Rucio, provided they cannot be found in
        the local cache.
        :param uninjectedData: same data as it's returned from the uninjectedFiles
        :return: set of containers successfully inserted into Rucio
        """
        logging.info("Preparing to insert containers into Rucio...")
        newContainers = set()
        for location in uninjectedData:
            for container in uninjectedData[location]:
                # same container can be at multiple locations
                if container not in self.containersCache and container not in newContainers:
                    if self.rucio.createContainer(container, meta=self.metaDIDProject):
                        logging.info("Container %s inserted into Rucio", container)
                        newContainers.add(container)
                    else:
                        logging.error("Failed to create container: %s", container)
        logging.info("Successfully inserted %d containers into Rucio", newContainers)
        return newContainers

    def insertBlocks(self, uninjectedData):
        """
        This method will insert blocks into Rucio and attach them to their correspondent
        containers, when attaching this block, we also need to provide the RSE that it
        will be available.
        :param uninjectedData: same data as it's returned from the uninjectedFiles
        :return: a dictionary of successfully inserted blocks and their correspondent location
        """
        logging.info("Preparing to insert blocks into Rucio...")
        newBlocks = set()
        for location in uninjectedData:
            rseName = "%s_Test" % location if self.testRSEs else location
            for container in uninjectedData[location]:
                for block in uninjectedData[location][container]:
                    if block not in self.blocksCache:
                        if self.rucio.createBlock(block, rse=rseName, meta=self.metaDIDProject):
                            logging.info("Block %s inserted into Rucio", block)
                            newBlocks.add(block)
                        else:
                            logging.error("Failed to create block: %s", block)
        logging.info("Successfully inserted %d blocks into Rucio", newBlocks)
        return newBlocks

    def insertBlockRules(self):
        """
        Creates a simple replication rule for every single block that
        is under production in a given site/RSE.
        Also persist the rule ID in the database.
        """
        if not self.createBlockRules:
            return

        logging.info("Preparing to create block rules into Rucio...")

        unsubBlocks = self.getUnsubscribedBlocks.execute()

        for item in unsubBlocks:
            # first, check if the block has already been created in Rucio
            if not self.rucio.didExist(item['blockname']):
                logging.warning("Block: %s not yet in Rucio. Retrying later..", item['blockname'])
                continue
            kwargs = dict(activity="Production Output", account=self.rucioAcct,
                          grouping="DATASET", comment="WMAgent automatic container rule",
                          ignore_availability=True, meta=self.metaData)
            rseName = "%s_Test" % item['pnn'] if self.testRSEs else item['pnn']
            # DATASET = replicates all files in the same block to the same RSE
            kwargs.update(self.blockRuleParams)
            resp = self.rucio.createReplicationRule(item['blockname'],
                                                    rseExpression=rseName, **kwargs)
            if resp:
                msg = "Block rule created for block: %s, at: %s, with rule id: %s"
                logging.info(msg, item['blockname'], item['pnn'], resp[0])
                binds = {'RULE_ID': resp[0], 'BLOCKNAME': item['blockname']}
                self.setBlockRules.execute(binds)
            else:
                logging.error("Failed to create rule for block: %s at %s", item['blockname'], rseName)
        return

    def insertReplicas(self, uninjectedData):
        """
        Inserts replicas into Rucio and attach them to its specific block.
        If the insertion succeeds, also switch their database state to injected.

        :param uninjectedData: dictionary with blocks as key, and RSEs as value
        """
        # FIXME: I think we need a different data struct from the database
        # this method is very expensive O(n^4)
        logging.info("Preparing to insert replicas into Rucio...")

        for location in uninjectedData:
            rseName = "%s_Test" % location if self.testRSEs else location
            for container in uninjectedData[location]:
                for block in uninjectedData[location][container]:
                    if block not in self.blocksCache:
                        logging.warning("Skipping %d file injection for block that failed to be added into Rucio: %s",
                                        len(uninjectedData[location][container][block]['files']), block)
                        continue
                    injectData = []
                    listLfns = []
                    for fileInfo in uninjectedData[location][container][block]['files']:
                        listLfns.append(fileInfo['lfn'])
                        injectData.append(dict(name=fileInfo['lfn'], scope=self.scope,
                                               bytes=fileInfo['size'], state="A",
                                               adler32=fileInfo['checksum']['adler32']))

                    if self.rucio.createReplicas(rse=rseName, files=injectData, block=block):
                        logging.info("Successfully inserted %d files on block %s", len(listLfns), block)
                        self._updateLFNState(listLfns)
        return

    def _updateLFNState(self, listLfns, recovery=False):
        """
        Given a list of LFNs, update their state in dbsbuffer table.
        :param listLfns: list of LFNs
        :param recovery: True if we are recovering previously injected files
        :return: nothing
        """
        if not listLfns:
            return

        try:
            self.setStatus.execute(listLfns, 1)
        except Exception as ex:
            # save it to try to inject them again in the next cycle
            self.filesToRecover.extend(listLfns)
            if 'Deadlock found' in str(ex) or 'deadlock detected' in str(ex):
                logging.error("Deadlock during file status update. Retrying again in the next cycle.")
                self.filesToRecover.extend(listLfns)
            else:
                msg = "Failed to update file status in the database, reason: %s" % str(ex)
                logging.error(msg)
                raise RucioInjectorException(msg) from None
        else:
            if recovery:
                self.filesToRecover = []

    def closeBlocks(self):
        """
        Close any blocks that have been migrated to global DBS
        """
        logging.info("Starting closeBlocks method")

        # in short, dbsbuffer_file.in_phedex = 1 AND dbsbuffer_block.status = 'InDBS'
        migratedBlocks = self.getMigrated.execute()
        # FIXME the data format returned by this DAO
        for location in migratedBlocks:
            for container in migratedBlocks[location]:
                for block in migratedBlocks[location][container]:
                    logging.info("Closing block: %s", block)
                    if self.rucio.closeBlockContainer(block):
                        self.setBlockClosed.execute(block)
                    else:
                        logging.error("Failed to close block: %s. Will retry again later.", block)

    def deleteBlocks(self):
        """
        _deleteBlocks_
        Find deletable blocks, then decide if to delete based on:
        Has transfer to all destinations finished ?
          If yes => Delete rules associated with the block, set deleted=1
          If no => do nothing (check again next cycle)
        """
        logging.info("Checking if there are block rules to be deleted...")

        # Get list of blocks that can be deleted
        # blockDict = self.findDeletableBlocks.execute(transaction=False)
        completedBlocksDict = self.getCompletedBlocks.execute(transaction=False)

        if not completedBlocksDict:
            logging.info("No candidate blocks found for rule deletion.")
            return

        logging.info("Found %d completed blocks", len(completedBlocksDict))
        logging.debug("Full completedBlocksDict: %s", pformat(completedBlocksDict))

        deletableWfsDict = set(self.getDeletableWorkflows.execute())

        if not deletableWfsDict:
            logging.info("No workflow chains (Parent + child workflows) in fully terminal state found. Skipping block level rule deletion in the current run.")
            return

        logging.info("Found %d workflows in terminal state.", len(deletableWfsDict))
        logging.debug("Full deletableWfsDict: %s", pformat(deletableWfsDict))

        now = int(time.time())
        blockDict = {}
        for block in completedBlocksDict.values():
            if block['workflowNames'].issubset(deletableWfsDict) and \
               now - block['blockCreateTime'] > self.blockDeletionDelaySeconds:
                blockDict[block['blockName']] = block

        logging.info("Found %d final candidate blocks for rule deletion", len(blockDict))
        logging.debug("Final deletable blocks dict: %s", pformat(blockDict))

        for blocksSlice in grouper(blockDict, self.delBlockSlicesize):
            logging.info("Handling %d candidate blocks", len(blocksSlice))
            containerDict = {}
            # Populate containerDict, assigning each block to its correspondent container
            for blockName in blocksSlice:
                container = blockDict[blockName]['dataset']
                # If the container is not in the dictionary, create a new entry for it
                if container not in containerDict:
                    # All blocks belonging to a container need to be sent to the same sites, so we simply take the sites list 
                    # from the current block to determine the containers required final RSEs.
                    sites = set(x.replace("_MSS", "_Tape") for x in blockDict[blockName]['sites'])
                    containerDict[container] = {'blocks': [], 'rse': sites}
                containerDict[container]['blocks'].append(blockName)

            blocksToDelete = []
            for contName in containerDict:
                cont = containerDict[contName]

                # Checks if the container is not requested in any sites.
                # This should never be triggered, but better safe than sorry
                if not cont['rse']:
                    logging.warning("No rules for container: %s. Its blocks won't be deleted.", contName)
                    continue

                try:
                    # Get RSE in which each block is available
                    availableRSEs = self.rucio.getReplicaInfoForBlocks(block=cont['blocks'], deep=self.useDsetReplicaDeep)
                except Exception as exc:
                    msg = "Failed to get replica info for blocks in container: %s.\n" % contName
                    msg += "Will retry again in the next cycle. Error: %s" % str(exc)
                    logging.error(msg)
                    continue

                for blockRSEs in availableRSEs:
                    # If block is available at every RSE its container needs to be transferred, the block can be deleted
                    blockSites = set(blockRSEs['replica'])
                    logging.debug("BlockName: %s", blockRSEs['name'])
                    logging.debug("Needed: %s / Available: %s", str(cont['rse']), str(blockSites))
                    if cont['rse'].issubset(blockSites):
                        blocksToDelete.append(blockRSEs['name'])

            # Delete agent created rules locking the block
            binds = []
            logging.info("Going to delete %d block rules", len(blocksToDelete))
            for block in blocksToDelete:
                try:
                    rules = self.rucio.listDataRules(block, scope=self.scope, account=self.rucioAcct)
                except WMRucioException:
                    logging.warning("Unable to retrieve replication rules for block: %s. Will retry in the next cycle.", block)
                else:
                    if not rules:
                        logging.info("Block rule for: %s has been deleted by previous cycles", block)
                        binds.append({'DELETED': 1, 'BLOCKNAME': block})
                        continue
                    for rule in rules:
                        deletedRules = 0
                        if self.rucio.deleteRule(rule['id'], purgeReplicas=True):
                            logging.info("Successfully deleted rule: %s, for block %s.", rule['id'], block)
                            deletedRules += 1
                        else:
                            logging.warning("Failed to delete rule: %s, for block %s. Will retry in the next cycle.", rule['id'], block)
                    if deletedRules == len(rules):
                        binds.append({'DELETED': 1, 'BLOCKNAME': block})
                        logging.info("Successfully deleted all rules for block %s.", block)

            self.markBlocksDeleted.execute(binds)
            logging.info("Marked %d blocks as deleted in the database", len(binds))

        return

    def insertContainerRules(self):
        """
        Polls the database for containers meant to be subscribed and create
        a container level rule to replicate all the files to a given RSE.
        It deals with both Central Production and T0 data rules, which require
        a different approach, such as:
          * Production Tape/Custodial data placement is skipped and data is marked as transferred
          * Production Disk/NonCutodial has a generic RSE expression and some rules override
            from the agent configuration (like number of copies, grouping and weight)
          * T0 Tape is created as defined, with a special rule activity for Tape
          * T0 Disk is created as defined, with a special rule activity for Disk/Export
        """
        logging.info("Starting insertContainerRules method")

        ruleComment = "WMAgent automatic container rule"
        if self.isT0agent:
            ruleComment = "T0 " + ruleComment

        # FIXME also adapt the format returned by this DAO
        # Check for completely unsubscribed datasets that are already marked as in_phedex = 1
        unsubscribedDatasets = self.getUnsubscribedDsets.execute()

        # Keep a list of subscriptions to tick as subscribed in the database
        subscriptionsMade = []

        # Create the subscription objects and add them to the list
        # The list takes care of the sorting internally
        for subInfo in unsubscribedDatasets:
            rseName = subInfo['site'].replace("_MSS", "_Tape")
            container = subInfo['path']
            lifetime = subInfo['dataset_lifetime']
            rulepriority = RUCIO_RULES_PRIORITY.get(subInfo['priority'])
            # Skip central production Tape rules
            if not self.isT0agent and rseName.endswith("_Tape"):
                logging.info("Bypassing Production container Tape data placement for container: %s and RSE: %s",
                             container, rseName)
                subscriptionsMade.append(subInfo['id'])
                continue
            # then check if the container has already been created in Rucio
            if not self.rucio.didExist(container):
                logging.warning("Container: %s not yet in Rucio. Retrying later..", container)
                continue

            ruleKwargs = dict(ask_approval=False,
                              activity=self._activityMap(rseName),
                              account=self.rucioAcct,
                              grouping="ALL",
                              comment=ruleComment,
                              priority=rulepriority,
                              meta=self.metaData)
            if not rseName.endswith(("_Tape", "_Tape_Test")):
                # add extra parameters to the Disk rule as defined in the component configuration
                ruleKwargs.update(self.containerDiskRuleParams)

            if not self.isT0agent:
                # destination for production Disk rules are always overwritten
                rseName = self.containerDiskRuleRSEExpr
                if self.testRSEs:
                    rseName = rseName.replace("cms_type=real", "cms_type=test")
            else:
                # then it's a T0 container placement
                if not rseName.endswith("_Tape") and lifetime > 0:
                    ruleKwargs['lifetime'] = lifetime
                if self.testRSEs:
                    rseName = "%s_Test" % rseName
                # Checking whether we need to ask for rule approval
                try:
                    if self.rucio.requiresApproval(rseName):
                        ruleKwargs['ask_approval'] = True
                except WMRucioException as exc:
                    msg = str(exc)
                    msg += "\nUnable to check approval requirements. Will retry again in the next cycle."
                    logging.error(msg)
                    continue

            logging.info("Creating container rule for %s against RSE %s", container, rseName)
            logging.debug("Container rule will be created with keyword args: %s", ruleKwargs)
            try:
                resp = self.rucio.createReplicationRule(container,
                                                        rseExpression=rseName, **ruleKwargs)
            except Exception:
                msg = "Failed to create container rule for (retrying with approval): %s" % container
                logging.warning(msg)
                ruleKwargs["ask_approval"] = True
                try:
                    resp = self.rucio.createReplicationRule(container,
                                                            rseExpression=rseName, **ruleKwargs)
                except Exception as exc:
                    msg = "Failed once again to create container rule for: %s " % container
                    msg += "\nWill retry again in the next cycle. Error: %s" % str(exc)
                    continue
            if resp:
                logging.info("Container rule created for %s under rule id: %s", container, resp)
                subscriptionsMade.append(subInfo['id'])
            else:
                logging.error("Failed to create rule for container: %s", container)

        # Register the result in DBSBuffer
        if subscriptionsMade:
            self.markSubscribed.execute(subscriptionsMade)
            logging.info("%d containers successfully locked in Rucio and local database", len(subscriptionsMade))

        return

    def _activityMap(self, rseName):
        """
        It maps the WMAgent type (Production vs T0) and the RSE name to
        properly set the rule activity field
        :param rseName: a string with the RSE name
        :return: a string with the rule activity
        """
        if not self.isT0agent and not rseName.endswith("_Tape"):
            return "Production Output"
        elif self.isT0agent and rseName.endswith(("_Tape", "_Tape_Test")):
            return "T0 Tape"
        elif self.isT0agent:
            return "T0 Export"
        else:
            msg = "This code should never be reached. Report it to the developers. "
            msg += "Trying to create container rule for RSE name: {}".format(rseName)
            raise WMRucioException(msg)



#!/usr/bin/env python
"""
_Delete_

MySQL implementation of Locations.Delete
"""

__all__ = []



from WMCore.Database.DBFormatter import DBFormatter

class Delete(DBFormatter):
    sql = "delete from wmbs_location where site_name = :location"

    def getBinds(self, location = None):
        return self.dbi.buildbinds(self.dbi.makelist(location), 'location')

    def execute(self, siteName = None, conn = None, transaction = False):
        self.dbi.processData(self.sql, self.getBinds(location = siteName),
                         conn = conn, transaction = transaction)
        return True



"""
Set of rules and functions
Handling resubmission Block construction
"""

from builtins import object

ACDC_PREFIX = "acdc"

class ACDCBlock(object):

    @staticmethod
    def name(wmspecName, taskName, offset, numOfFile):
        taskName = taskName.replace('/', ':')
        return "/%s/%s/%s/%s/%s" % (ACDC_PREFIX, wmspecName, taskName, offset,
                                     numOfFile)

    @staticmethod
    def checkBlockName(blockName):
        if blockName.split('/')[1] == ACDC_PREFIX:
            return True
        else:
            return False

    @staticmethod
    def splitBlockName(blockName):
        """ return False if the blockName is not acdc Block
            return original elements of block name in order as a dict
            {'SpecName': blockSplit[1],
             'TaskName': blockSplit[2],
             'Offset': int(blockSplit[3]),
             'NumOfFiles': int(blockSplit[4])},
             if it is acdc Block
            raise ValueError if the block is has wrong format
        """
        blockSplit = blockName.split('/')
        if blockSplit[1] != ACDC_PREFIX:
            return False
        elif len(blockSplit) != 6:
            msg = """blockName should contain prefix, wmspec name, task name,
                     offset and number of files %s""" % blockName
            raise ValueError(msg)
        else:
            return {'SpecName': str(blockSplit[2]),
                    'TaskName': str(blockSplit[3]).replace(':', '/'),
                    'Offset': int(blockSplit[4]),
                    'NumOfFiles': int(blockSplit[5])}



"""
_CountFinishedSubscriptionsByWorkflow_

MySQL implementation of Subscription.CountFinishedSubscriptionsByWorkflow
"""
from __future__ import division, print_function

from builtins import str, bytes

from WMCore.Database.DBFormatter import DBFormatter


class CountFinishedSubscriptionsByWorkflow(DBFormatter):
    """
    Gets count subscriptions given workflow but exculding cleanup and log collect subscription
    """

    sql = """SELECT ww.name as workflow,
             COUNT(case when ws.finished = 1 then 1 else null end) as finished,
             COUNT(case when ws.finished = 0 then 1 else null end) as open
             FROM wmbs_subscription ws
             INNER JOIN wmbs_sub_types wst ON wst.id = ws.subtype
             INNER JOIN wmbs_workflow ww ON ww.id = ws.workflow
             WHERE wst.name NOT IN ('LogCollect', 'Cleanup') AND ww.name = :workflowName
             GROUP BY ww.name
          """

    def execute(self, workflowNames, conn=None, transaction=False):
        """
        _execute_

        This DAO returns a list of dictionaries containing
        the key 'id' with the id of the finished subscriptions
        """
        if isinstance(workflowNames, (str, bytes)):
            workflowNames = [workflowNames]

        binds = [{'workflowName': w} for w in workflowNames]

        if not binds:
            return []

        result = self.dbi.processData(self.sql, binds, conn=conn, transaction=transaction)
        return self.formatDict(result)



#!/usr/bin/env python
"""
_GetFilesForParentlessMerge_

MySQL implementation of Subscription.GetFilesForParentlessMerge
"""

from WMCore.Database.DBFormatter import DBFormatter


class GetFilesForParentlessMerge(DBFormatter):
    """
    This query needs to return the following for any files that is deemed
    mergeable:
      WMBS ID (file_id)
      Events (file_events)
      Size (file_size)
      LFN (file_lfn)
      First event in file (file_first_event)
      Runs in file (file_run)
      Lumi sections in file (file_lumi)
      PNN
    """

    sql = """SELECT wmbs_file_details.id AS file_id,
                    wmbs_file_details.events AS file_events,
                    wmbs_file_details.filesize AS file_size,
                    wmbs_file_details.lfn AS file_lfn,
                    wmbs_file_details.first_event AS file_first_event,
                    MIN(wmbs_file_runlumi_map.run) AS file_run,
                    MIN(wmbs_file_runlumi_map.lumi) AS file_lumi,
                    wmbs_pnns.pnn,
                    wmbs_fileset_files.insert_time AS insert_time,
                    wmbs_workflow.injected AS injected
             FROM wmbs_sub_files_available
               INNER JOIN wmbs_file_details ON
                 wmbs_sub_files_available.fileid = wmbs_file_details.id
               INNER JOIN wmbs_file_runlumi_map ON
                 wmbs_file_details.id = wmbs_file_runlumi_map.fileid
               INNER JOIN wmbs_file_location ON
                 wmbs_file_details.id = wmbs_file_location.fileid
               INNER JOIN wmbs_pnns ON
                 wmbs_file_location.pnn = wmbs_pnns.id
               INNER JOIN wmbs_subscription ON
                 wmbs_subscription.id = wmbs_sub_files_available.subscription
               INNER JOIN wmbs_fileset_files ON
                 wmbs_fileset_files.fileset = wmbs_subscription.fileset AND
                 wmbs_fileset_files.fileid = wmbs_sub_files_available.fileid
               INNER JOIN wmbs_workflow ON
                 wmbs_workflow.id = wmbs_subscription.workflow
             WHERE wmbs_sub_files_available.subscription = :p_1
             GROUP BY wmbs_file_details.id,
                      wmbs_file_details.events,
                      wmbs_file_details.filesize,
                      wmbs_file_details.lfn,
                      wmbs_file_details.first_event,
                      wmbs_pnns.pnn,
                      wmbs_fileset_files.insert_time,
                      wmbs_workflow.injected
             """

    def execute(self, subscription=None, conn=None, transaction=False):
        results = self.dbi.processData(self.sql, {"p_1": subscription}, conn=conn,
                                       transaction=transaction)
        return self.formatDict(results)



#!/usr/bin/env python
"""
_ArhciveDataReporterr_

"""
from __future__ import (division, print_function)
import logging
import threading

from WMCore.Agent.Harness import Harness
from WMComponent.ArchiveDataReporter.ArchiveDataPoller import ArchiveDataPoller

class ArchiveDataReporter(Harness):
    """
    Component class for AgentStatusWatcher module
    """

    def __init__(self, config):
        """
        __init__

        Initialize the Harness
        """
        Harness.__init__(self, config)

        return

    def preInitialization(self):
        """
        _preInitialization_

        Sets up the worker thread
        """
        logging.info("ArhciveDataReporter.preInitialization")
        pollInterval = self.config.ArchiveDataReporter.ArchiveDataPoller.pollInterval
        myThread = threading.currentThread()

        if not self.config.ArchiveDataReporter.WMArchiveURL:
            logging.info("No archive url is set: Doing nothing")
        else:
            logging.info("Setting ResourcesUpdate poll interval to %s seconds" % pollInterval)
            myThread.workerThreadManager.addWorker(ArchiveDataPoller(self.config), pollInterval)

        return



#!/usr/bin/env python
"""
_BasicCounter_

Count jobs created so far for a task and insert the count into the job

"""
from WMCore.JobSplitting.Generators.GeneratorInterface import GeneratorInterface



class BasicCounter(GeneratorInterface):
    """
    _BasicCounter_

    Incremental job counter

    TODO: Use WMAgent couch to get count for the particular task as starting value in ctor
          Or use options for better control

    """
    def __init__(self, **options):
        GeneratorInterface.__init__(self, **options)
        self.count = 0

    def __call__(self, wmbsJob):


        wmbsJob['counter'] = self.count
        self.count += 1



#!/usr/bin/env python
"""
_GetFWJRByState_

Oracle implementation of Jobs.GetFWJRByState
"""




from WMCore.WMBS.MySQL.Jobs.GetFWJRByState import GetFWJRByState as MySQLGetFWJRByState

class GetFWJRByState(MySQLGetFWJRByState):
    """
    Identical to MySQL version.
    """
    pass



#!/usr/bin/env python
"""
Generate.py

A class that parses a workflow specification file
(or agent specification file) and generates the appropriate
stub classes.

"""
from __future__ import print_function





from builtins import str, object
from future.utils import viewvalues

import os
import sys
import pickle

from WMCore.Agent.Configuration import loadConfigurationFile


class Generate(object):
    """
    Generate

    A class that parses a workflow specification file
    (or agent specification file) and generates the appropriate
    stub classes.
    """


    def __init__(self, configFile):
        self.config = loadConfigurationFile(configFile)
        # current dir when writing stub files.
        self.currentDir = None

        # parsed component information from configuration file.
        self.components = {}
        # parsed synchronizers (trigger) information from configuration file.
        self.synchronizers = {}
        # default stub message that is put in every generated file.
        self.stubmsg = """
\"\"\"
Auto generated stub be careful with editing,
Inheritance is preferred.
\"\"\"


"""
    def generate(self):
        """
        Main method to be called, parses the file and generates the classes.
        """

        try:
            print('Creating output directory '+\
                str(self.config.General.srcDir))
            os.makedirs(self.config.General.srcDir)
        except:
            print('ERROR: Make sure directory does not exist already')
            sys.exit(0)
        self.parse()
        print('Starting generation')
        self.componentStubs()
        self.componentTests()
        print('Flow generated')

    def parse(self):
        """
        Parses the input file.
        """

        print('Parsing config file')
        # preprocess
        for synchronizer in self.config.General.synchronizers:
            synchro = pickle.loads(synchronizer)
            self.synchronizers[synchro['ID']] = synchro
            self.synchronizers[synchro['ID']]['components'] = []
        for handler in self.config.General.handlers:
            hndlr = pickle.loads(handler)
            if hndlr['component'] not in self.components:
                self.components[hndlr['component']] = {}
            self.components[hndlr['component']][hndlr['messageIn']] = hndlr
            if 'synchronize' in hndlr:
                self.synchronizers[hndlr['synchronize']]['components'].\
                    append(hndlr['component'])
        for plugin in self.config.General.plugins:
            plgn = pickle.loads(plugin)
            self.components[plgn['component']][plgn['handler']]['plugin'] = \
                 plgn


    def componentTests(self):
        msg = """
class %sTest(unittest.TestCase):

    _setup_done = False

    def setUp(self):

        if not %sTest._setup_done:
            self.testInit = TestInit(__file__)
            self.testInit.setLogging()
            self.testInit.setDatabaseConnection()
            %sTest._setup_done = True

    def testA(self):
        config = self.testInit.getConfiguration(\
            os.path.join(os.getenv('APPBASE'), \
            '%s/%s/DefaultConfig.py'))
        theComponent = %s(config)
        # make sure database settings are set properly.
        #theComponent.prepareToStart()
        print('I am a placeholder for a test')

    def runTest(self):

        #Run test.
        self.testA()

if __name__ == '__main__':
    unittest.main()
"""
        for componentName in self.components:
            print('Creating test dir for:'+componentName)
            self.currentDir = os.path.join(self.config.General.testDir, \
                componentName+'_t')
            os.makedirs(self.currentDir)
            print('Creating test stub')
            with open(os.path.join(self.currentDir,'__init__.py'), 'w') as stfile:
                stfile.write('#!/usr/bin/env python\n')

            with open(os.path.join(self.currentDir, componentName + '_t.py'), 'w') as stfile:
                stfile.write('#!/usr/bin/env python\n')
                stfile.write(self.stubmsg)
                stfile.write("# test file skeletons.\n\n\n")
                stfile.write('import os\n')
                stfile.write('import unittest\n\n\n')
                stfile.write("from "+self.config.General.pythonPrefix+'.'+componentName+'.'+componentName+' import '+componentName+'\n')
                stfile.write("from WMCore.Agent.Configuration import Configuration\n")
                stfile.write("from WMQuality.TestInit import TestInit\n")
                stfile.write(msg % (componentName, componentName, componentName,
                                    self.config.General.srcDir, componentName, componentName))

        self.currentDir = os.path.join(self.config.General.baseDir, 'standards')
        try:
            os.makedirs(self.currentDir)
        except:
            pass
        print('Creating test suite')
        with open(os.path.join(self.currentDir,'defaultTest.py'), 'w') as stfile:
            stfile.write("#!/usr/bin/env python\n")
            stfile.write("from WMQuality.Test import Test\n\n\n")
            for componentName in self.components:
                msg = "from "+self.config.General.pythonTestPrefix+"."+componentName+'_t.'+componentName+'_t import '+componentName+'Test'
                stfile.write(msg+'\n')
            stfile.write('\n\n\n')
            stfile.write('errors = {}\n')
            stfile.write('tests = []\n')

        msg1 = """

try:
   x=%sTest()
   tests.append((x,"nobody"))
except Exception,ex:
   if not errors.has_key("nobody"):
       errors["nobody"] = []
   errors["nobody"].append(("%sTest",str(ex)))

"""
        msg2 = """
print('Writing level 2 failures to file: failures2.log ')
failures = open('failures2.log','w')

failures.writelines('Failed instantiation summary (level 2): \\\n')
for author in errors.keys():
    failures.writelines('\\\n*****Author: '+author+'********\\\n')
    for errorInstance, errorMsg in  errors[author]:
        failures.writelines('Test: '+errorInstance)
        failures.writelines(errorMsg)
        failures.writelines('\\\n\\\n')
failures.close()

test = Test(tests,'failures3.log')
test.run()
test.summaryText()
"""
        for componentName in self.components:
            stfile.write(msg1 % (componentName, componentName))
            stfile.write(msg2)

    def componentStubs(self):
        """
        Generates the component stubs.
        """

        for componentName in self.components:
            print('Creating component dir for:'+componentName)
            self.currentDir = os.path.join(self.config.General.srcDir, \
                componentName)
            os.makedirs(self.currentDir)
            print('Creating component stub')
            with open(os.path.join(self.currentDir,'__init__.py'), 'w') as stfile:
                stfile.write('#!/usr/bin/env python\n')

            stfile = \
                open(os.path.join(self.currentDir, componentName + '.py'), 'w')
            stfile.write('#!/usr/bin/env python\n')
            stfile.write(self.stubmsg)
            stfile.write("# harness class that encapsulates the basic component logic.\n")
            stfile.write("from WMCore.Agent.Harness import Harness\n\n")
            importFact = False
            try:
                for handlerName in self.components[componentName]:
                    handler = self.components[componentName][handlerName]
                    if 'configurable' in handler and handler['configurable'] == 'yes':
                        if not importFact:
                            stfile.write("# we do not import handler " + self.convert(handlerName) + " as they are dynamicly\n")
                            stfile.write("# loaded from the config file.\n")
                            stfile.write("from WMCore.WMFactory import WMFactory\n")
                            importFact = True
                    else:
                        stfile.write('from ' + self.config.General.pythonPrefix + '.' )
                        stfile.write(componentName +'.Handler.' + self.convert(handlerName) +' import')
                        stfile.write(' ' + self.convert(handlerName)+'\n')
            except Exception as ex:
                print('No messages for component: '+ componentName)
            msg = """
class %s(Harness):


    def __init__(self, config):
        # call the base class
        Harness.__init__(self, config)


    def preInitialization(self):
    # mapping from message types to handlers
""" % (componentName)
            stfile.write(msg)

            if importFact:
                stfile.write('        # use a factory to dynamically load handlers.\n')
                stfile.write("        factory = WMFactory('generic')\n")
            try:
                for handlerName in self.components[componentName]:
                    handler = self.components[componentName][handlerName]
                    if 'configurable' in handler and \
                        handler['configurable'] == 'yes':
                        msg = """
        # in case nothing was configured we have a fallback.
        if not hasattr(self.config.%s, "%sHandler"):
            self.config.%s.%sHandler =\\
            '%s.%s.Handler.%s'
        self.messages['%s'] = \\
            factory.loadObject(\\
            self.config.%s.%sHandler, self)
""" % (componentName, self.convert(handlerName), componentName, self.convert(handlerName), \
      self.config.General.pythonPrefix, componentName, self.convert(handlerName), \
      handlerName, componentName, self.convert(handlerName))
                        stfile.write(msg)
                    else:
                        stfile.write("        self.messages['"+handlerName+"'] = "+ self.convert(handlerName)+"(self)\n\n")
            except Exception as ex:
                print('is this an ERROR? :'+str(ex))
            stfile.flush()
            stfile.close()
            print('Creating handler stubs')
            self.defaultConfig(componentName)
            self.handlerStubs(componentName)

    def defaultConfig(self, componentName):
        """
        Generates the default config file.
        """

        stfile = open(os.path.join(self.currentDir, 'DefaultConfig.py'), 'w')
        stfile.write('#!/usr/bin/env python\n')
        stfile.write(self.stubmsg)
        msg = """
from WMCore.Agent.Configuration import Configuration

config = Configuration()
config.component_("%s")
#The log level of the component.
config.%s.logLevel = "INFO"
""" % (componentName, componentName)
        stfile.write(msg)
        # check if we need a thread parameter
        threadParam = False
        for handler in viewvalues(self.components[componentName]):
            if 'threading' in handler and handler['threading'] == 'yes':
                threadParam = True
        if threadParam:
            stfile.write('# maximum number of threads we want to deal\n')
            stfile.write('# with messages per pool.\n')
            stfile.write('config.'+componentName+'.maxThreads = 30\n')
        for handlerName in self.components[componentName]:
            handler = self.components[componentName][handlerName]
            if 'configurable' in handler and handler['configurable'] == 'yes':
                stfile.write('config.'+componentName+'.'+self.convert(handlerName)+'Handler =\\\n')
                stfile.write('    "'+self.config.General.pythonPrefix+'.'+componentName+'.Handler.'+self.convert(handlerName)+'"\n')

        stfile.flush()
        stfile.close()


    def handlerStubs(self, componentName):
        """
        Generates the handler stub files.
        """

        handlerDir = os.path.join(self.currentDir, 'Handler')
        try:
            os.makedirs(handlerDir)
        except:
            pass
        with open(os.path.join(handlerDir, '__init__.py'), 'w') as stfile:
            stfile.write('#!/usr/bin/env python\n')
        for handlerName in self.components[componentName]:
            # check if we need to create a threaded version
            handler = self.components[componentName][handlerName]
            threaded = False
            if 'threading' in handler:
                if handler['threading'] == 'yes':
                    threaded = True

            stfile = open(os.path.join(handlerDir, self.convert(handlerName) +'.py'),'w')
            stfile.write('#!/usr/bin/env python\n')
            stfile.write(self.stubmsg)
            if not threaded:
                stfile.write('import threading\n\n')
            stfile.write('from WMCore.Agent.BaseHandler import BaseHandler\n')
            if threaded:
                stfile.write('from WMCore.ThreadPool.ThreadPool import ThreadPool\n')
            msg = """


class %s(BaseHandler):

    def __init__(self, component):
        BaseHandler.__init__(self, component)

""" % (self.convert(handlerName))
            stfile.write(msg)
            if threaded:
                msg = """
        self.threadpool = ThreadPool(\\
            "%s.%s.Handler.%sSlave",\\
            self.component, "%s",\\
            self.component.config.%s.maxThreads)

""" %(self.config.General.pythonPrefix, componentName, \
                    self.convert(handlerName), self.convert(handlerName), componentName)
                stfile.write(msg)

            msg = """
     # this we overload from the base handler
    def __call__(self, event, payload):
        \"\"\"
        Implement the handler here.
        Assign values to "messageToBePublished","yourTaskId"
        ,"yourPayloadString,"yourActionPayload"".
        Where necessary
        \"\"\"
"""
            stfile.write(msg)
            if not threaded:
                msg = """
        print("Implement me: %s.%s.Handler.%s")

        myThread = threading.currentThread()

""" % (self.config.General.pythonPrefix, \
                    componentName, self.convert(handlerName))
                stfile.write(msg)
                self.triggers(stfile, componentName, handlerName)
                self.messages(stfile, componentName, handlerName)
            else:
                stfile.write('        self.threadpool.enqueue(event, {"event" : event, "payload" : payload})\n')
                self.threadSlave(componentName, handlerName)

            stfile.flush()
            stfile.close()

    def threadSlave(self, componentName, handlerName):
        """
        Generates thread slave stubs.
        """

        slaveDir = os.path.join(self.currentDir, 'Handler')
        className = self.convert(handlerName) + 'Slave'
        with open(os.path.join(slaveDir, className + '.py'),'w') as stfile:
            stfile.write('#!/usr/bin/env python\n')
            stfile.write(self.stubmsg)
            stfile.write('import threading\n')
            stfile.write('#inherit from our default slave implementation\n')
            stfile.write('from WMCore.ThreadPool.ThreadSlave import ThreadSlave\n')
            stfile.write('\n')
            stfile.write('class '+className + '(ThreadSlave):\n')
            stfile.write('\n')
            stfile.write('    def __call__(self, parameters):\n')
            stfile.write('        """\n')
            stfile.write('        Implement the handler here.\n')
            stfile.write('        Assign values to "messageToBePublished","yourTaskId"\n')
            stfile.write('        ,"yourPayloadString,"yourActionPayload"".\n')
            stfile.write('        Where necessary\n')
            stfile.write('        """\n')
            stfile.write('\n')
            stfile.write('        print("Implement me:')
            stfile.write(self.config.General.pythonPrefix+'.'+componentName+'.Handler.'+self.convert(handlerName) + '")')
            stfile.write('\n\n')
            stfile.write('        myThread = threading.currentThread()\n\n')
            self.triggers(stfile, componentName, handlerName)
            self.messages(stfile, componentName, handlerName)
            stfile.write('\n\n')
            stfile.write('        # we need to do this in our slave otherwise the \n')
            stfile.write('        # messages that might have been published, will not be send.\n')
            stfile.write('        myThread.msgService.finish()\n')

    def messages(self, stfile, componentName, handlerName):
        """
        Generates message statements in handler, slave.
        """
        # check what messages need to be published and how (e.g. choice or no choice)
        messagesOut = self.components[componentName][handlerName]['messageOut']
        if messagesOut.find('|') > -1 and messagesOut != '':
            messagesOut = messagesOut.split('|')
            for messageOut in messagesOut:
                stfile.write("        if messageToBePublished == '")
                stfile.write(messageOut+ "':\n")
                msg = "            msg = {'name':'"+messageOut+"', 'payload': yourPayloadString}"
                stfile.write(msg+'\n')
                stfile.write("            myThread.msgService.publish(msg)\n")
        elif messagesOut != '':
            messagesOut = messagesOut.split(',')
            for messageOut in messagesOut:
                msg = "        msg = {'name':'"+messageOut+"', 'payload': yourPayloadString}"
                stfile.write(msg+'\n')
                stfile.write("        myThread.msgService.publish(msg)\n")

    def triggers(self, stfile, componentName, handlerName):
        """
        Generates trigger statements.
        """

        handler = self.components[componentName][handlerName]

        if 'createSynchronizer' in handler:
            stfile.write('        flags = []\n')
            for synchronizer in self.synchronizers[handler['createSynchronizer']]['components']:
                flag = "{'trigger_id' : '"+handler['createSynchronizer']+ "',\\\n" + \
                        "                'id' : yourTaskId,\\\n"+ \
                        "                'flag_id' : '"+synchronizer+"'}"
                stfile.write('        flag = '+flag+'\n')
                stfile.write('        flags.append(flag)\n')
            stfile.write('        myThread.trigger.addFlag(flags)\n')
            action = "{'trigger_id' : '"+ handler['createSynchronizer']+ "',\\\n" + \
                     "                  'id' : yourTaskId,\\\n"+ \
                     "                  'action_name' : '"+self.synchronizers[handler['createSynchronizer']]['action']+"',\\\n" +\
                     "                  'payload' : yourActionPayload}"
            stfile.write('        action = '+action+'\n')
            stfile.write('        myThread.trigger.setAction(action)\n')
        if 'synchronize' in self.components[componentName][handlerName]:
            flag = "{'trigger_id' : '"+handler['synchronize']+ "'," + \
                    "'id' : yourTaskId,"+ \
                    "'flag_id' : '"+componentName+"'}"
            stfile.write('        flag = '+flag+'\n')
            stfile.write('        myThread.trigger.addFlag(flags)\n')

    def convert(self, textStr):
        """
        filters out unwanted symbols excluded in python
        class names
        """
        no = [':','/']

        for symbol in no:
            textStr = textStr.replace(symbol, '_')
        return textStr



#!/usr/bin/env python
"""
_Parentage_

Oracle implementation of Fileset.New

"""
__all__ = []



from WMCore.WMBS.MySQL.Fileset.New import New as NewFilesetMySQL

class New(NewFilesetMySQL):
    sql = """INSERT INTO wmbs_fileset (name, last_update, open)
             VALUES (:NAME, :LAST_UPDATE, :OPEN)"""



#!/usr/bin/env python
"""
_LoadBulkFilesByID_

Oracle implementation of DBSBufferFiles.LoadBulkFilesByID
"""

from WMComponent.DBS3Buffer.MySQL.DBSBufferFiles.LoadBulkFilesByID import LoadBulkFilesByID as MySQLLoadBulkFilesByID

class LoadBulkFilesByID(MySQLLoadBulkFilesByID):
    """
    Same as MySQL

    """
    pass



#!/usr/bin/env python
"""
_MySQLCore_

Handle bind variable parsing for MySQL.
"""

import copy

from WMCore.Database.DBCore import DBInterface
from WMCore.Database.ResultSet import ResultSet

def bindVarCompare(a):
    """
    _bindVarCompare_

    Bind variables are represented as a tuple with the first element being the
    variable name and the second being it's position in the query.  We sort on
    the position in the query.
    """
    return a[1]

def stringLengthCompare(a):
    """
    _stringLengthCompare_

    Sort comparison function to sort strings by length.
    Since we want to sort from longest to shortest, this must be reversed when used
    """
    return len(a)


class MySQLInterface(DBInterface):
    def substitute(self, origSQL, origBindsList):
        """
        _substitute_

        Transform as set of bind variables from a list of dictionaries to a list
        of tuples:

        b = [ {'bind1':'value1a', 'bind2': 'value2a'},
        {'bind1':'value1b', 'bind2': 'value2b'} ]

        Will be transformed into:

        b = [ ('value1a', 'value2a'), ('value1b', 'value2b')]

        Don't need to substitute in the binds as executemany does that
        internally. But the sql will also need to be reformatted, such that
        :bind_name becomes %s.

        See: http://www.devshed.com/c/a/Python/MySQL-Connectivity-With-Python/5/
        """
        if origBindsList == None:
            return origSQL, None

        origBindsList = self.makelist(origBindsList)
        origBind = origBindsList[0]

        bindVarPositionList = []
        updatedSQL = copy.copy(origSQL)

        # We process bind variables from longest to shortest to avoid a shorter
        # bind variable matching a longer one.  For example if we have two bind
        # variables: RELEASE_VERSION and RELEASE_VERSION_ID the former will
        # match against the latter, causing problems.  We'll sort the variable
        # names by length to guard against this.
        bindVarNames = list(origBind)
        bindVarNames.sort(key=stringLengthCompare, reverse=True)

        bindPositions = {}
        for bindName in bindVarNames:
            searchPosition = 0

            while True:
                bindPosition = origSQL.lower().find(":%s" % bindName.lower(),
                                                    searchPosition)
                if bindPosition == -1:
                    break

                if bindPosition not in bindPositions:
                    bindPositions[bindPosition] = 0
                    bindVarPositionList.append((bindName, bindPosition))
                searchPosition = bindPosition + 1

            searchPosition = 0
            while True:
                bindPosition = updatedSQL.lower().find(":%s" % bindName.lower(),
                                                       searchPosition)

                if bindPosition == -1:
                    break

                left = updatedSQL[0:bindPosition]
                right = updatedSQL[bindPosition + len(bindName) + 1:]
                updatedSQL = left + "%s" + right

        bindVarPositionList.sort(key=bindVarCompare)

        mySQLBindVarsList = []
        for origBind in origBindsList:
            mySQLBindVars = []
            for bindVarPosition in bindVarPositionList:
                mySQLBindVars.append(origBind[bindVarPosition[0]])

            mySQLBindVarsList.append(tuple(mySQLBindVars))

        return (updatedSQL, mySQLBindVarsList)

    def executebinds(self, s = None, b = None, connection = None,
                     returnCursor = False):
        """
        _executebinds_

        Execute a SQL statement that has a single set of bind variables.
        Transform the bind variables into the format that MySQL expects.
        """
        s, b = self.substitute(s, b)
        return DBInterface.executebinds(self, s, b, connection, returnCursor)

    def executemanybinds(self, s = None, b = None, connection = None,
                         returnCursor = False):
        """
        _executemanybinds_

        Execute a SQL statement that has multiple sets of bind variables.
        Transform the bind variables into the format that MySQL expects.
        """
        newsql, binds = self.substitute(s, b)

        return DBInterface.executemanybinds(self, newsql, binds, connection,
                                            returnCursor)



#!/usr/bin/env python
"""
_AutoIncrementCheck_

AutoIncrement Check
Test to properly set the autoIncrement value

First, find the highest jobID either in wmbs_job or in wmbs_highest_job
Then reset AUTO_INCREMENT to point to that.
"""

__all__ = []

import logging

from WMCore.Database.DBFormatter import DBFormatter

class AutoIncrementCheck(DBFormatter):
    """
    _AutoIncrmentCheck_

    Check and properly set the auto_increment counter for wmbs_job
    """


    highestSQL = """SELECT IFNULL(MAX(id), 0) FROM wmbs_job"""

    currentSQL = """SELECT Auto_increment FROM information_schema.tables WHERE table_name='wmbs_job' AND table_schema=DATABASE()"""

    alterSQL = "ALTER TABLE wmbs_job AUTO_INCREMENT = :value"


    def execute(self, input = 0, conn = None, transaction = False):
        """
        _execute_


        """

        highest = self.dbi.processData(self.highestSQL, {}, conn = conn,
                                       transaction = transaction)[0].fetchall()[0][0]

        current = self.dbi.processData(self.currentSQL, {}, conn = conn,
                                       transaction = transaction)[0].fetchall()[0][0]

        value = max(input + 1, highest + 1)

        if value > current:
            self.dbi.processData(self.alterSQL, {'value': value},
                                 conn = conn, transaction = transaction)

        return



from WMCore.WMException import WMException

class WMSpecFactoryException(WMException):
    """
    _WMSpecFactoryException_

    This exception will be raised by validation functions if
    the code fails validation.  It will then be changed into
    a proper HTTPError in the ReqMgr, with the message you enter
    used as the message for farther up the line.
    """
    pass



#!/usr/bin/env python
"""
_LoadFromIDWithType_

Oracle implementation of LoadFromIDWithType

Created on Oct 3, 2012

@author: dballest
"""

from WMCore.WMBS.MySQL.Jobs.LoadFromIDWithType import LoadFromIDWithType as MySQLLoadFromIDWithType

class LoadFromIDWithType(MySQLLoadFromIDWithType):
    pass



#!/usr/bin/env python
"""
_StageOutError_

General Exception class for JC modules

"""




import inspect
import sys

from WMCore.WMException import WMException

ErrorDefinitions = {
    60311 : "GeneralStageOutFailure",
    60315 : "StageOutInitError",
    }


class StageOutError(WMException):
    """
    _StageOutError_

    Exception class which works out details of where
    it was raised.

    """
    def __init__(self, message, **data):
        WMException.__init__(self, message, **data)

        self.data.setdefault("ErrorCode", 60311)
        self.data.setdefault("ErrorType", ErrorDefinitions[60311])

class StageOutFailure(StageOutError):
    """
    _StageOutFailure_

    Standard Error class to indicate a stage out failure

    """
    def __init__(self, message, **data):
        StageOutError.__init__(self, message, **data)
        self.data.setdefault("ErrorCode", 60311)
        self.data.setdefault("ErrorType", ErrorDefinitions[60311])

        return

class StageOutInitError(StageOutError):
    """
    _StageOutInitError_

    Error when initialising for stage out including reading
    site conf, TFC etc

    """
    def __init__(self, message, **data):
        StageOutError.__init__(self, message, **data)
        self.data["ErrorCode"] =  60315
        self.data["ErrorType"] = ErrorDefinitions[60315]

class StageOutInvalidPath(StageOutError):
    """
    Exception class indicating a directory does not exist
    """
    def __init__(self, messgae = '', **data):
        StageOutError.__init__(self, messgae, **data)



"""
ArchiveDataPoller
"""

from __future__ import (division, print_function)
import logging
import traceback
from Utils.IteratorTools import grouper
from Utils.Timers import timeFunction
from Utils.wmcoreDTools import resetWatchdogTimer, moduleName
from WMCore.WorkerThreads.BaseWorkerThread import BaseWorkerThread
from WMCore.Services.WMArchive.DataMap import createArchiverDoc
from WMCore.Services.WMArchive.WMArchive import WMArchive
from WMCore.Services.FWJRDB.FWJRDBAPI import FWJRDBAPI


class ArchiveDataPoller(BaseWorkerThread):
    """
    Gather fwjr data and update to archiver\
    """

    def __init__(self, config):
        """
        initialize properties specified from config
        """
        BaseWorkerThread.__init__(self)
        self.config = config

    def setup(self, parameters):
        """
        set db connection(couchdb, wmbs) to prepare to gather information
        """
        baseURL = self.config.JobStateMachine.couchurl
        dbname = "%s/fwjrs" % getattr(self.config.JobStateMachine, "couchDBName")

        self.fwjrAPI = FWJRDBAPI(baseURL, dbname)
        self.wmarchiver = WMArchive(self.config.ArchiveDataReporter.WMArchiveURL)
        self.numDocsRetrievePerPolling = getattr(self.config.ArchiveDataReporter, "numDocsRetrievePerPolling", 1000)
        self.numDocsUploadPerCall = getattr(self.config.ArchiveDataReporter, "numDocsUploadPerCall", 200)

    @timeFunction
    def algorithm(self, parameters):
        """
        get information from wmbs, workqueue and local couch
        """
        try:
            data = self.fwjrAPI.getFWJRByArchiveStatus('ready', limit=self.numDocsRetrievePerPolling)['rows']
            logging.info("Found %i not archived documents from FWRJ db to upload to WMArchive.", len(data))

            for slicedData in grouper(data, self.numDocsUploadPerCall):
                jobIDs = []
                archiveDocs = []
                for job in slicedData:
                    doc = createArchiverDoc(job)
                    archiveDocs.append(doc)
                    jobIDs.append(job["id"])

                response = self.wmarchiver.archiveData(archiveDocs)

                # Partial success is not allowed either all the insert is successful or none is
                if response[0]['status'] == "ok" and len(response[0]['ids']) == len(jobIDs):
                    archiveIDs = response[0]['ids']
                    for docID in jobIDs:
                        self.fwjrAPI.updateArchiveUploadedStatus(docID)
                    logging.info("...successfully uploaded %d docs", len(jobIDs))
                    logging.debug("JobIDs uploaded: %s", jobIDs)
                    logging.debug("Archived IDs returned: %s", archiveIDs)
                else:
                    logging.warning("Upload failed and it will be retried in the next cycle: %s: %s.",
                                    response[0]['status'], response[0]['reason'])
                    logging.debug("failed JobIds %s", jobIDs)
        except Exception as ex:
            logging.error("Error occurred, will retry later:")
            logging.error(str(ex))
            logging.error("Trace back: \n%s", traceback.format_exc())

        # Reset its own watchdog timer at the end of the run cycle
        logging.info(f"Resetting {moduleName(self)} watchdog timer.")
        if resetWatchdogTimer(self):
            logging.warning(f"Failed to reset {moduleName(self)} watchdog timer. The component might be restarted soon.")



#!/usr/bin/env python
"""
_Generators_

Generator implementations to be passed to a JobFactory instance in order
to generate custom quantities for jobs outside of the normal
WMBS Job Splitting quantities

"""
__all__ = []



#!/usr/bin/env python
"""
_New_

MySQL implementation of JobGroup.SetSite
"""

__all__ = []



from WMCore.Database.DBFormatter import DBFormatter

class SetSite(DBFormatter):
    sql = """UPDATE wmbs_jobgroup jg
              SET jg.location = (SELECT ID FROM wmbs_location WHERE site_name = :site_name)
              WHERE jg.ID = :jobGroupID"""

    def execute(self, site_name = None, jobGroupID = None, conn = None,
                transaction = False):
        binds = {'site_name' : site_name, 'jobGroupID' : jobGroupID}
        result = self.dbi.processData(self.sql, binds, conn = conn,
                                      transaction = transaction)
        return



"""
Unlike ReqMgr1 defining Request and RequestSchema classes,
define just 1 class. Derived from Python dict and implementing
necessary conversion and validation extra methods possibly needed.


    since arguments validation #4705, arguments which are later
        validated during spec instantiation and which are not
        present in the request injection request, can't be defined
        here because their None value is not allowed in the spec.
        This is the case for e.g. DbsUrl, AcquisitionEra
        This module should probably define only absolutely
        necessary request parameters and not any optional ones.

"""
from __future__ import print_function, division
from builtins import range, object
from future.utils import viewitems, viewvalues, listvalues

import re
import time
from copy import deepcopy
from WMCore.REST.Auth import get_user_info
from WMCore.ReqMgr.DataStructs.RequestStatus import REQUEST_START_STATE, ACTIVE_STATUS_FILTER


def initialize_request_args(request, config):
    """
    Request data class request is a dictionary representing
    a being injected / created request. This method initializes
    various request fields. This should be the ONLY method to
    manipulate request arguments upon injection so that various
    levels or arguments manipulation does not occur across several
    modules and across about 7 various methods like in ReqMgr1.

    request is changed here.
    """

    # user information for cert. (which is converted to cherry py log in)
    user = get_user_info()
    request["Requestor"] = user["login"]
    request["RequestorDN"] = user.get("dn", "unknown")
    # service certificates carry @hostname, remove it if it exists
    request["Requestor"] = request["Requestor"].split('@')[0]

    # assign first starting status, should be 'new'
    request["RequestStatus"] = REQUEST_START_STATE
    request["RequestTransition"] = [{"Status": request["RequestStatus"],
                                     "UpdateTime": int(time.time()), "DN": request["RequestorDN"]}]
    request["RequestDate"] = list(time.gmtime()[:6])

    # set the original priority when request is create
    request["PriorityTransition"] = [{"Priority": request["RequestPriority"],
                                      "UpdateTime": int(time.time()), "DN": request["RequestorDN"]}]
    # update the information from config
    request["CouchURL"] = config.couch_host
    request["CouchWorkloadDBName"] = config.couch_reqmgr_db
    request["CouchDBName"] = config.couch_config_cache_db

    generateRequestName(request)
    replaceDbsProdUrl(request)


def _replace_cloned_args(clone_args, user_args):
    """
    replace original arguments with user argument.
    If the value is dictionary format, overwrite only with specified arguments.
    If the original argument has simple value and user passes dictionary, completely replace to dictionary
    XXX this means LumiList won't remove other runs, only updates runs overwritten
    """
    for prop in user_args:
        if isinstance(user_args[prop], dict) and isinstance(clone_args.get(prop), dict):
            _replace_cloned_args(clone_args.get(prop, {}), user_args[prop])
        else:
            clone_args[prop] = user_args[prop]
    return


def replaceDbsProdUrl(requestArgs):
    """
    Function to update the DBS URL from the standard cmsweb.cern.ch
    to cmsweb-prod.cern.ch, running in the k8s infra
    :param requestArgs: dictionary with the request arguments
    :return: same dictionary object (with updated dbs url)
    """
    # TODO: this url conversion below can be removed in one year from now, thus March 2022
    if requestArgs.get("DbsUrl"):
        requestArgs["DbsUrl"] = requestArgs["DbsUrl"].replace("cmsweb.cern.ch", "cmsweb-prod.cern.ch")
        requestArgs["DbsUrl"] = requestArgs["DbsUrl"].rstrip("/")


def initialize_clone(requestArgs, originalArgs, argsDefinition, chainDefinition=None):
    """
    Initialize arguments for a clone request by inheriting and overwriting argument
    from OriginalRequest.

    :param requestArgs: user-provided dictionary with override arguments
    :param originalArgs: original arguments retrieved for the workflow being cloned
    :param argsDefinition: arguments definition according to the workflow type being cloned
    :param chainDefinition: a dictionary containing the chain argument definition, for
    StepChain and TaskChain
    :return: dictionary with original args filtered out, as per the spec definition. And on
     top of that, user arguments added/replaced in the dictionary.
    """
    chainDefinition = chainDefinition or {}
    chainPattern = r'(Task|Step)\d{1,2}'

    if originalArgs == argsDefinition:
        # then it's a clone/ACDC of another ACDC, nothing else to do
        cloneArgs = originalArgs
    else:
        cloneArgs = {}
        for topKey, topValue in viewitems(originalArgs):
            # order of this if-else matters because Step1/Task1 is a known argument
            if re.match(chainPattern, topKey):
                cloneArgs.setdefault(topKey, {})
                # remove unsupported keys from inner Step/Task dict
                for innerKey in topValue:
                    if innerKey in chainDefinition:
                        cloneArgs[topKey][innerKey] = topValue[innerKey]
            # accepts floating Skim args for ReReco
            elif topKey in argsDefinition or topKey.startswith('Skim'):
                cloneArgs[topKey] = topValue

    # apply user override arguments at the end, such that it's validated at spec level
    incrementProcVer(cloneArgs, requestArgs)
    _replace_cloned_args(cloneArgs, requestArgs)
    replaceDbsProdUrl(cloneArgs)

    return cloneArgs


def incrementProcVer(cloneArgs, requestArgs):
    """
    Increment the ProcessingVersion value for any requests cloned via
    clone API, except if it's a Resubmission request.
    TODO: ProcVer can be a dict at top level, until this #6881 gets fixed
    """
    # if either the parent or the new workflow is a Resubmission, we shall not increment ProcVer
    if cloneArgs['RequestType'] == 'Resubmission' or requestArgs.get('RequestType') == 'Resubmission':
        return
    for key in cloneArgs:
        if key == 'ProcessingVersion':
            if isinstance(cloneArgs[key], int):
                cloneArgs[key] += 1
            elif isinstance(cloneArgs[key], dict):
                for taskname in cloneArgs[key]:
                    cloneArgs[key][taskname] += 1
        elif isinstance(cloneArgs[key], dict) and 'ProcessingVersion' in cloneArgs[key]:
            cloneArgs[key]['ProcessingVersion'] += 1
    return


def generateRequestName(request):
    currentTime = time.strftime('%y%m%d_%H%M%S', time.localtime(time.time()))
    seconds = int(10000 * (time.time() % 1.0))

    request["RequestName"] = "%s_%s" % (request["Requestor"], request.get("RequestString"))
    request["RequestName"] += "_%s_%s" % (currentTime, seconds)


def protectedLFNs(requestInfo):
    reqData = RequestInfo(requestInfo)
    result = []
    if reqData.andFilterCheck(ACTIVE_STATUS_FILTER):
        outs = requestInfo.get('OutputDatasets', [])
        base = requestInfo.get('UnmergedLFNBase', '/store/unmerged')
        for out in outs:
            dsn, ps, tier = out.split('/')[1:]
            acq, rest = ps.split('-', 1)
            dirPath = '/'.join([base, acq, dsn, tier, rest])
            result.append(dirPath)
    return result


class RequestInfo(object):
    """
    Wrapper class for Request data
    """

    def __init__(self, requestData):
        self.data = requestData

    def _maskTaskStepChain(self, prop, chain_name, default=None):

        propExist = False
        numLoop = self.data["%sChain" % chain_name]
        for i in range(numLoop):
            if prop in self.data["%s%s" % (chain_name, i + 1)]:
                propExist = True
                break

        defaultValue = self.data.get(prop, default)

        if propExist:
            result = set()
            for i in range(numLoop):
                chain_key = "%s%s" % (chain_name, i + 1)
                chain = self.data[chain_key]
                if prop in chain:
                    foundValue = chain[prop]
                else:
                    if isinstance(defaultValue, dict):
                        foundValue = defaultValue.get(chain_key, None)
                    else:
                        foundValue = deepcopy(defaultValue)

                if foundValue not in [None, ""]:
                    if isinstance(foundValue, (list, set)):
                        result.update(foundValue)
                    else:
                        result.add(foundValue)
            return list(result)
        else:
            # property which can't be task or stepchain property but in dictionary format
            exculdePropWithDictFormat = ["LumiList", "AgentJobInfo"]
            if prop not in exculdePropWithDictFormat and isinstance(defaultValue, dict):
                return listvalues(defaultValue)
            else:
                return defaultValue

    def get(self, prop, default=None):
        """
        gets the value when prop exist as one of the properties in the request document.
        In case TaskChain, StepChain workflow it searches the property in Task/Step level
        """

        if "TaskChain" in self.data:
            return self._maskTaskStepChain(prop, "Task", default)
        elif "StepChain" in self.data:
            return self._maskTaskStepChain(prop, "Step", default)
        elif prop in self.data:
            return self.data[prop]
        else:
            return default

    def andFilterCheck(self, filterDict):
        """
        checks whether filterDict condition met.
        filterDict is the dict of key and value(list) format)
        i.e.
        {"RequestStatus": ["running-closed", "completed"],}
        If this request's RequestStatus is either "running-closed", "completed",
        return True, otherwise False
        """
        for key, value in viewitems(filterDict):
            # special case checks where key is not exist in Request's Doc.
            # It is used whether AgentJobInfo is deleted or not for announced status
            if value == "CLEANED" and key == "AgentJobInfo":
                if self.isWorkflowCleaned():
                    continue
                else:
                    return False

            if isinstance(value, dict):
                # TODO: need to handle dictionary comparison
                # For now ignore
                continue
            elif value in ["false", "False", "FALSE"]:
                value = False
            elif value in ["true", "True", "TRUE"]:
                value = True

            # Now make value list if value is not in list form including bool value
            if not isinstance(value, list):
                value = [value]

            reqValue = self.get(key)
            if reqValue is not None:
                if isinstance(reqValue, list):
                    if not set(reqValue).intersection(set(value)):
                        return False
                elif reqValue not in value:
                    return False
            else:
                return False
        return True

    def isWorkflowCleaned(self):
        """
        check whether workflow data is cleaned up from agent only checks the couchdb
        Since dbsbuffer data is not clean up we can't just check 'AgentJobInfo' key existence
        This all is only meaningfull if request status is right before end status.
        ["aborted-completed", "rejected", "announced"]
        DO NOT check if workflow status isn't among those status
        """
        if 'AgentJobInfo' in self.data:
            for agentRequestInfo in viewvalues(self.data['AgentJobInfo']):
                if agentRequestInfo.get("status", {}):
                    return False
        # cannot determine whether AgentJobInfo is cleaned or not when 'AgentJobInfo' key doesn't exist
        # Maybe JobInformation is not included but since it requested by above status assumed it returns True
        return True



#!/usr/bin/env python
"""
_SetLocationByLFN_

Oracle implementation of DBSBuffer.SetLocationByLFN
"""

from WMComponent.DBS3Buffer.MySQL.DBSBufferFiles.SetLocationByLFN import SetLocationByLFN as MySQLSetLocationByLFN


class SetLocationByLFN(MySQLSetLocationByLFN):
    """
    Set the location of files using lfn as the key

    """
    sql = """INSERT /*+ IGNORE_ROW_ON_DUPKEY_INDEX (dbsbuffer_file_location (filename, location)) */
               INTO dbsbuffer_file_location (filename, location)
             SELECT df.id, dl.id FROM dbsbuffer_file df, dbsbuffer_location dl
             WHERE df.lfn = :lfn AND dl.pnn = :pnn
          """



#!/usr/bin/env python
"""
_ListJobsBySub_

Retrieve a list of all jobs generated by a particular subscription.  For each
job return the couch record, outcome, state and name.  The response is a list
of dictionaries with each dictionary having the following keys:
    couch_record
    outcome
    jobstate
    jobname
"""




from WMCore.WMBS.MySQL.Monitoring.DefaultFormatter import DefaultFormatter

class ListJobsBySub(DefaultFormatter):
    sql = """SELECT wmbs_job.name AS jobName, wmbs_job_state.name as jobState,
                    wmbs_job.couch_record, wmbs_job.outcome FROM wmbs_job
               INNER JOIN wmbs_jobgroup ON
                 wmbs_job.jobgroup = wmbs_jobgroup.id
               INNER JOIN wmbs_job_state ON
                 wmbs_job.state = wmbs_job_state.id
              WHERE wmbs_jobgroup.subscription = :subscriptionId"""

    def execute(self, subscriptionId, conn = None, transaction = False):
        """
        _execute_

        Execute the SQL for the given job ID and then format and return
        the result.
        """
        result = self.dbi.processData(self.sql, {"subscriptionId": subscriptionId},
                                      conn = conn, transaction = transaction)

        return self.formatDict(result)
